#include "uda1380.h"
#include "lpc177x_8x_i2c.h"
#include "lpc177x_8x_pinsel.h"
#include "lpc177x_8x_timer.h"

//Uda1380 link to I2C0 only
#define UDA1380_I2C    I2C_0

struct uda1380_setup
{
  uint8_t reg;
  uint16_t val;
};


#define UDA1380_END    0xFF
#define UDA1380_DELAY  0xDD
const struct uda1380_setup UDA1380_InitPara[]= {
  // Restore L3 default values (software reset)
  {UDA1380_REG_L3,              0x00FF},
  // Enable all power for now
  {UDA1380_REG_PWRCTRL,         0xA5DF},
  // CODEC ADC and DAC clock from WSPLL, all clocks enabled
  {UDA1380_REG_EVALCLK,         0x0332},
  // I2S bus data I/O formats, use digital mixer for output BCKO is slave
  {UDA1380_REG_I2S,             0x0000},
   // Full mixer analog input gain
  {UDA1380_REG_ANAMIX,          0x2222},
  // Enable headphone short circuit protection
  {UDA1380_REG_HEADAMP,         0x0202},

  {UDA1380_DELAY,               0x0064}, //100ms

  // Full master volume
  {UDA1380_REG_MSTRVOL,         0x4040},
  // Enable full mixer volume on both channels
  {UDA1380_REG_MIXVOL,          0xFF00},
  // Bass and treble boost set to flat
  {UDA1380_REG_MODEBBT,         0x0000},
  // Disable mute and de-emphasis
  {UDA1380_REG_MSTRMUTE,        0x0000},
  // Mixer off, other settings off
  {UDA1380_REG_MIXSDO,          0x0000},
  // ADC decimator volume to max
  {UDA1380_REG_DECVOL,          0x0000},
   // No PGA mute, full gain
  {UDA1380_REG_PGA,             0x0000},
  // Select line in and MIC, max MIC gain
  {UDA1380_REG_ADC,             0x0F02},
  // AGC
  {UDA1380_REG_AGC,             0x0000},

  // End of list
  {UDA1380_END,                 0xFFFF}
};

/*********************************************************************//**
 * @brief     Initialize Uda1380
 * @param[in]  i2cClockFreq  I2C clock frequency that Uda1380 operate
 * @param[in] i2sClockFreq  I2S bit clock frequency
 * @return     UDA1380_FUNC_ERR or UDA1380_FUNC_OK
 **********************************************************************/
int32_t Uda1380_Init(uint32_t i2cClockFreq, uint32_t i2sClockFreq)
{
  const struct uda1380_setup *preset;
  
  // Config Pin for SYSCLK
  PINSEL_ConfigPin (1, 16, 2);

  // Config Pin for I2C_SDA and I2C_SCL of I2C0
  // It's because the Uda1380 IC is linked to LPC177x_8x by I2C0 clearly
  PINSEL_ConfigPin (0, 27, 1);
  PINSEL_ConfigPin (0, 28, 1);

  I2C_Init(UDA1380_I2C, i2cClockFreq);

  /* Enable I2C operation */
  I2C_Cmd(UDA1380_I2C, ENABLE);

  /* Init*/
  preset = UDA1380_InitPara;

  while (preset->reg != UDA1380_END)
  {
    if (preset->reg == UDA1380_DELAY)
    {
      TIM_Waitms(preset->val);
    }
    else
    {
      Uda1380_WriteData(preset->reg, preset->val);
    }
    preset++;
  }
  
  return UDA1380_FUNC_OK;
}

/*********************************************************************//**
 * @brief     Write data to a register of Uda1380
 * @param[in]  reg  Register address
 * @param[out] data  data value.
 * @return     UDA1380_FUNC_ERR or UDA1380_FUNC_OK
 **********************************************************************/
int32_t Uda1380_WriteData(uint8_t reg, uint16_t data)
{
  I2C_M_SETUP_Type i2cData;
  uint8_t i2cBuf[UDA1380_CMD_BUFF_SIZE];

  i2cBuf[0] = reg;
  i2cBuf[1] = (data >> 8) & 0xFF;
  i2cBuf[2] = data & 0xFF;

  i2cData.sl_addr7bit = UDA1380_SLAVE_ADDR;
  i2cData.tx_length = UDA1380_CMD_BUFF_SIZE;
  i2cData.tx_data = i2cBuf;
  i2cData.rx_data = NULL;
  i2cData.rx_length = 0;
  i2cData.retransmissions_max = 3;

  if (I2C_MasterTransferData(UDA1380_I2C, &i2cData, I2C_TRANSFER_POLLING) == SUCCESS)
  {
      uint16_t dataTmp;
      if(Uda1380_ReadData(reg, &dataTmp) != UDA1380_FUNC_OK) {
      return UDA1380_FUNC_ERR;
      }
    if(dataTmp != data)
      return UDA1380_FUNC_ERR;

    return UDA1380_FUNC_OK;
  }

  return UDA1380_FUNC_ERR;
}


/*********************************************************************//**
 * @brief     Read data stored in register of Uda1380
 * @param[in]  reg  Register address
 * @param[out] data  point to the buffer which is used for storing data.
 * @return     UDA1380_FUNC_ERR or UDA1380_FUNC_OK
 **********************************************************************/
int32_t Uda1380_ReadData(uint8_t reg, uint16_t *data)
{
  I2C_M_SETUP_Type i2cData;
  uint8_t i2cBuf[UDA1380_CMD_BUFF_SIZE];

  if(data == NULL)
    return UDA1380_FUNC_ERR;

  i2cBuf[0] = reg;

  i2cData.sl_addr7bit = UDA1380_SLAVE_ADDR;
  i2cData.tx_length = 1;
  i2cData.tx_data = i2cBuf;
  i2cData.rx_data = &i2cBuf[1];
  i2cData.rx_length = UDA1380_CMD_BUFF_SIZE - 1;
  i2cData.retransmissions_max = 3;

  if (I2C_MasterTransferData(UDA1380_I2C, &i2cData, I2C_TRANSFER_POLLING) == SUCCESS)
  {
    *data = i2cBuf[1] << 8 | i2cBuf[2];
    return UDA1380_FUNC_OK;
  }

  return UDA1380_FUNC_ERR;
}

/*********************************************************************//**
 * @brief       Control the master volume.
 * @param[in]  left  volume in left channel
 * @param[in]  right  volume in right channel
 * @return     UDA1380_FUNC_ERR or UDA1380_FUNC_OK
 **********************************************************************/
int32_t Uda1380_MasterVolCtrl (uint8_t left, uint8_t right)
{
  return Uda1380_WriteData(UDA1380_REG_MSTRVOL, left | (right<<8));
}

