/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "lpc_types.h"
#include "lpc177x_8x_timer.h"
#include "tsc2046_touch.h"
#include "calibrate.h"
#include "lpc_swim_font.h"
#include "lpc_swim.h"
#include "globals.h"
#include "application.h"
#include "trans_fs.h"
#include "lpc_heap.h"
#include "pca9532.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define BACK_COLOR WHITE
#define FORE_COLOR BLACK

/*
 * Allows user to hold down SW2 button during boot to recalibrate touch
 */
#define FORCED_CALIBRATION() (pca9532_getLedState(FALSE) & KEY1)

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static Bool firstTime = TRUE;

static int calibPoint = 0;

static tTouchPoint refPoint1 = { 15,  242};
static tTouchPoint refPoint2 = {240, 80};
static tTouchPoint refPoint3 = {460, 222};
static tTouchPoint scr1 = {0, 0};
static tTouchPoint scr2 = {0, 0};
static tTouchPoint scr3 = {0, 0};
static BOOL_32 releaseNeeded = FALSE;
static BOOL_32 touchReleased = FALSE;

static SWIM_WINDOW_T BaseWindow;

/******************************************************************************
 * Local functions
 *****************************************************************************/

static void
paintMessage(SWIM_WINDOW_T *win, char* pMsg)
{
  uint16_t msgLen = strlen((char *)pMsg);
  uint16_t x = 0;
  uint16_t y = 20;

  swim_set_pen_color(win, BLACK);
  swim_put_box(win, 0, 0, 480, 40);
  swim_set_pen_color(win, WHITE);

  // each character is 6 pixels wide
  x = (240 / 2) - (msgLen*6/2);

  swim_put_text_xy(win, (const CHAR*)pMsg, x, y);
}


static void
drawCalibPoint(SWIM_WINDOW_T *win, UNS_16 x, UNS_16 y)
{
  swim_clear_screen(win, BLACK);
  paintMessage(win, "Calibrating, touch marked box...");

  swim_put_box(win, x-5, y-5, x+5, y+5);

}

static void
calibrateStart(SWIM_WINDOW_T *win)
{
  INT_32 x, y, z = 0;

  BOOL_32 calibrated = FALSE;
  releaseNeeded = FALSE;
  calibPoint = 0;

  drawCalibPoint(win, refPoint1.x, refPoint1.y);

  touch_xyz(&x, &y, &z);

  while (!calibrated) {

    TIM_Waitms(100);

    touch_xyz(&x, &y, &z);

    if (z == 0)
    {
      touchReleased = TRUE;
      continue;
    }

    if (releaseNeeded && !touchReleased)
      continue;

    touchReleased = FALSE;


    switch (calibPoint++)
    {
    case 0:
      scr1.x = x;
      scr1.y = y;
      drawCalibPoint(win, refPoint2.x, refPoint2.y);

      releaseNeeded = TRUE;
      break;
    case 1:
      scr2.x = x;
      scr2.y = y;
      drawCalibPoint(win, refPoint3.x, refPoint3.y);

      releaseNeeded = TRUE;
      break;
    case 2:
      scr3.x = x;
      scr3.y = y;
      paintMessage(win, "Calibrated!");

      releaseNeeded = TRUE;
      touch_calibrate(refPoint1, refPoint2, refPoint3, scr1, scr2, scr3);

      calibrated = TRUE;
      break;

    }
  }
}

static Bool loadCalibrationFromFile(char* filename)
{
  tFsFileSystem fsType;
  uint8_t* pBuf;
  uint32_t size;
  int pos = 0;
  Bool calibrated = FALSE;

  if (transfs_loadFileIntoNewBuff(filename, ONLY_USE_FATFS, &pBuf, &size, &fsType) == TFS_OK) {
    if (size == 48) {
      uint32_t* pIntBuf = (uint32_t*)pBuf;
      refPoint1.x = pIntBuf[pos++];
      refPoint1.y = pIntBuf[pos++];
      refPoint2.x = pIntBuf[pos++];
      refPoint2.y = pIntBuf[pos++];
      refPoint3.x = pIntBuf[pos++];
      refPoint3.y = pIntBuf[pos++];
      scr1.x = pIntBuf[pos++];
      scr1.y = pIntBuf[pos++];
      scr2.x = pIntBuf[pos++];
      scr2.y = pIntBuf[pos++];
      scr3.x = pIntBuf[pos++];
      scr3.y = pIntBuf[pos++];

      touch_calibrate(refPoint1, refPoint2, refPoint3, scr1, scr2, scr3);
      calibrated = TRUE;
    }
    lpc_free(pBuf);
  }
  return calibrated;
}

static void saveCalibrationToFile(char* filename)
{
  uint32_t pIntBuff[12];
  pIntBuff[0] = refPoint1.x;
  pIntBuff[1] = refPoint1.y;
  pIntBuff[2] = refPoint2.x;
  pIntBuff[3] = refPoint2.y;
  pIntBuff[4] = refPoint3.x;
  pIntBuff[5] = refPoint3.y;
  pIntBuff[6] = scr1.x;
  pIntBuff[7] = scr1.y;
  pIntBuff[8] = scr2.x;
  pIntBuff[9] = scr2.y;
  pIntBuff[10] = scr3.x;
  pIntBuff[11] = scr3.y;

  transfs_storeBuffInFile(filename, TFS_FS_FAT, (uint8_t*)pIntBuff, 48);
}

static void runCalibration(void)
{
	// Adapt file name based on screen resolution so that the calibration
	// can be kept while switching between 4.3" and 7" displays.
	//
	// Filename will be /touch480.bin or /touch800.bin
	//
	char filename[14] = { '/', 't', 'o', 'u', 'c', 'h', '-', '-', '-', '.', 'b', 'i', 'n', '\0' };
  filename[6] = '0' + ((GlobalParameters.display.width/100)%10);
  filename[7] = '0' + ((GlobalParameters.display.width/10) %10);
  filename[8] = '0' + ( GlobalParameters.display.width     %10);
	
  // If this is the first time then attempt to load
  // settings from file system instead of calibration.
  // Future calls to runCalibration must calibrate.
  if (firstTime && !FORCED_CALIBRATION()) {
    firstTime = FALSE;
    if (loadCalibrationFromFile(filename)) {
      // Successfully loaded settings
      return;
    }
  }

  swim_window_open(&BaseWindow,
      GlobalParameters.display.width,
      GlobalParameters.display.height,
      (COLOR_T*)GlobalParameters.display.framebuffer, 0, 0,
      GlobalParameters.display.width-1,
      GlobalParameters.display.height-1,
      1, WHITE, BLACK, BLACK);

  // Make sure that the touch screen is not already calibrated
  touch_reinit();

  // Adjust calibration points to screen

  // Bottom left corner
  refPoint1.x = 15;
  refPoint1.y = GlobalParameters.display.height - 15;

  // Sort of top horizontal center
  refPoint2.x = GlobalParameters.display.width / 2;
  refPoint2.y = 80;

  // Bottom right corner
  refPoint3.x = GlobalParameters.display.width - 15;
  refPoint3.y = GlobalParameters.display.height - 15;

  calibrateStart(&BaseWindow);

  // Save data
  saveCalibrationToFile(filename);
}

static void cleanup(void)
{
  swim_window_close(&BaseWindow);
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/

static AppCallback_t calibrateCallback = {
    &runCalibration,
    &cleanup,
    NULL,
    0,
    FALSE,
};

AppCallback_t* calibAppInit(void)
{
  return &calibrateCallback;
}

