/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdio.h>
#include "lpc_types.h"
#include "application.h"
#include "lpc177x_8x_timer.h"
#include "tsc2046_touch.h"
#include "lpc_swim.h"
#include "lpc_swim_image.h"
#include "globals.h"
#include "log.h"
#include "image.h"
#include "swim_extensions.h"

/******************************************************************************
 * Defines, macros, and typedefs
 *****************************************************************************/

typedef struct {
  int oldX;
  int oldY;
  int oldZ;
  int newX;
  int newY;
  int newZ;
  Bool pressed;
} clickinfo_t;

typedef struct
{
  uint8_t* pData;
  uint16_t width;
  uint16_t height;
} image_t;

typedef enum
{
  MENU_EVENT_NOTHING,
  MENU_EVENT_CALIBRATE,
  MENU_EVENT_MP3_PLAYER,
  MENU_EVENT_3D_GRID,
  MENU_EVENT_3D_CUBE,
} menu_event_t;

typedef struct
{
	Bool inUse;
  int x0;
  int y0;
  int x1;
  int y1;
  menu_event_t action;
  char* descr;
} area_t;


#define NUM_AREAS (sizeof(clickable_areas)/sizeof(area_t))


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T MainWindow;

static clickinfo_t ClickInfo;

/*
 * Areas that the user may click on. All coordinates should be relative
 * to the image - not the display. The coordinates will be updated during
 * display initialization with the display's offset.
 */
static area_t clickable_areas[] = {
  {TRUE, 280,  42, 440,  77, MENU_EVENT_CALIBRATE,  "Calibrate"},       
  {TRUE, 280,  92, 440, 127, MENU_EVENT_MP3_PLAYER, "MP3"},       
  {TRUE, 280, 145, 440, 180, MENU_EVENT_3D_GRID,    "3D Grid"},       
  {TRUE, 280, 195, 440, 230, MENU_EVENT_3D_CUBE,    "3D Cube"},       
};


/******************************************************************************
 * Local functions
 *****************************************************************************/

static menu_event_t touchAt(int x, int y, int z, Bool released)
{
  //log_i("%s(%d, %d, %d) released = %d\r\n", __FUNCTION__, x, y, z, released);
  if (released)
  {
    int i;
    for (i = NUM_AREAS-1; i >= 0; i--)
    {
      if (clickable_areas[i].inUse     &&
				  (x >= clickable_areas[i].x0) && 
          (x <= clickable_areas[i].x1) && 
          (y >= clickable_areas[i].y0) && 
          (y <= clickable_areas[i].y1))
      {
        log_i("Clicked %s\r\n", clickable_areas[i].descr);
        return clickable_areas[i].action;
      }
    }
  }
  return MENU_EVENT_NOTHING;
}

static menu_event_t pollTouch(void)
{
  menu_event_t Event = MENU_EVENT_NOTHING;
  touch_xyz(&ClickInfo.newX, &ClickInfo.newY, &ClickInfo.newZ);

  if (ClickInfo.newZ == 0 && ClickInfo.oldZ != 0) 
	{
    // ignore x and y coordinates as they are wrong when z becomes 0
    ClickInfo.oldZ = 0;
    ClickInfo.pressed = FALSE;

    Event = touchAt(ClickInfo.oldX, ClickInfo.oldY, ClickInfo.oldZ, TRUE);
  } 
	else if ((ClickInfo.oldX != ClickInfo.newX) || 
		       (ClickInfo.oldY != ClickInfo.newY) || 
	         (ClickInfo.oldZ == 0 && ClickInfo.newZ != 0)) 
	{
    ClickInfo.oldX = ClickInfo.newX;
    ClickInfo.oldY = ClickInfo.newY;
    ClickInfo.oldZ = ClickInfo.newZ;
    if (!ClickInfo.pressed && ClickInfo.newZ > 0) 
		{
      ClickInfo.pressed = TRUE;
    }

    Event = touchAt(ClickInfo.oldX, ClickInfo.oldY, ClickInfo.oldZ, FALSE);
  }
  return Event;
}

static void updatePositions(int xoff, int yoff)
{
  int i;

  for (i = NUM_AREAS-1; i >= 0; i--)
  {
    clickable_areas[i].x0 += xoff;
    clickable_areas[i].x1 += xoff;
    clickable_areas[i].y0 += yoff;
    clickable_areas[i].y1 += yoff;
  }
}

static void showMenu(void)
{
  ImageError_t err;
	image_t imgMenu;
	menu_event_t event;

  // Menu background
  err = decodeImageFileRomFS(IMAGE_FORMAT_PNG, "/menu/menu.png", &imgMenu.pData, &imgMenu.width, &imgMenu.height);
  if (err != IMAGE_ERROR_OK) 
	{
    log_e("Failed to decode /menu.png, error %d\r\n", err);
  }
	else
	{
		int xoff, yoff;
		swim_window_open(&MainWindow,
				GlobalParameters.display.width,
				GlobalParameters.display.height,
				(COLOR_T*)GlobalParameters.display.framebuffer, 0, 0,
				GlobalParameters.display.width-1,
				GlobalParameters.display.height-1,
				0, WHITE, BLACK, BLACK);

		// Center the image
		xoff = (MainWindow.xpsize - imgMenu.width)/2;
		yoff = (MainWindow.ypsize - imgMenu.height)/2;
		swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgMenu.pData,
				xoff, yoff,	imgMenu.width, imgMenu.height);

		ClickInfo.oldX = ClickInfo.oldY = ClickInfo.oldZ = 0;
		ClickInfo.pressed = FALSE;
		
		updatePositions(xoff, yoff);
		
		do
		{
			do
			{
				event = pollTouch();
				TIM_Waitms(10);
			} while(event == MENU_EVENT_NOTHING);

			switch (event)
			{
				case MENU_EVENT_CALIBRATE:
					// Recalibrate
					appStartApp("Touch Screen Calibration");
					appRebootAfterDelay(1500);
					break;
				
				case MENU_EVENT_MP3_PLAYER:
					appStartApp("MP3 Player");
					break;
				
				case MENU_EVENT_3D_GRID:
					// Run visualization which will never return
					appStartApp("3D Grid Visualization");
					break;
				
				case MENU_EVENT_3D_CUBE:
					// Run visualization which will never return
					appStartApp("3D Cube Visualization");
					break;
			}
		} while(TRUE);
	}
}

static void cleanup(void)
{
	// Mandatory function, but it will never be called as the Menu system 
	// always runs
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/

static AppCallback_t menuCallback = {
    &showMenu,
    &cleanup,
    FALSE,
};

AppCallback_t* menuAppInit(void)
{
  return &menuCallback;
}
