/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdio.h>

#include "gui.h"
#include "lpc177x_8x_i2c.h"
#include "lpc177x_8x_timer.h"
#include "lpc177x_8x_wwdt.h"
#include "sdram.h"
#include "lpc_swim.h"
#include "lpc_swim_font.h"
#include "lpc_swim_image.h"
#include "ea_lcd_board.h"
#include "truly_display.h"
#include "tsc2046_touch.h"
#include "log.h"
#include "cfgmem.h"
#include "lpc_helvr10.h"
#include "lpc_heap.h"
#include "image.h"
#include "pca9532.h"
#include "globals.h"
#include "swim_extensions.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

typedef struct {
  uint16_t width;
  uint16_t height;
  uint16_t bkgOffsetX;
  uint16_t bkgOffsetY;
  uint16_t bkgWidth;
  uint16_t bkgHeight;
} displayinfo_t;

typedef struct {
  int oldX;
  int oldY;
  int oldZ;
  int newX;
  int newY;
  int newZ;
  Bool pressed;
} clickinfo_t;

typedef struct
{
  int min;
  int sec_high;
  int sec_low;
  int decimal;
} time_t;

typedef struct
{
  uint8_t* pData;
  uint16_t width;
  uint16_t height;
} image_t;

typedef struct
{
  int x;
  int y;
} point_t;

typedef struct
{
  Bool inUse;
  int x0;
  int y0;
  int x1;
  int y1;
  event_t action;
  char* descr;
} area_t;


#define I2C_PORT (I2C_0)

#define MAIN_BUF_SZ 100

#define CURSOR_MIN    30 
#define CURSOR_MAX    444
#define CURSOR_POS(_t,_m)  (((_m)>0) ? (((_t)*(CURSOR_MAX-CURSOR_MIN))/(_m)) : (0))

#define DIGIT_WIDTH  24
#define DIGIT_HEIGHT 35

#define DIGIT_IMG_WIDTH  DIGIT_WIDTH
#define DIGIT_IMG_HEIGHT 44
#define DIGIT_IMG_OFFSET(_dig) ((0 + (_dig)*DIGIT_IMG_HEIGHT)*DIGIT_IMG_WIDTH*2)

#define NUM_AREAS (sizeof(clickable_areas)/sizeof(area_t))


#define VU_BAR_WIDTH        7
#define VU_BAR_HEIGHT      32
#define VU_BAR_OFFSET_LX    3
#define VU_BAR_OFFSET_RX  (90 + VU_BAR_OFFSET_LX)
#define VU_BAR_OFFSET_Y    10
#define VU_BAR_SPACING    (3 + VU_BAR_WIDTH)
#define VU_LEVELS          32
#define VU_AREA_X0        280
#define VU_AREA_Y0        192
#define VU_AREA_WIDTH     176
#define VU_AREA_HEIGHT     50

/*
 * Support macros for extracting the R, G and B components (as uint8_t) from a 
 * RGB565 color in uint16_t and then to put it back together again.
 */
#define RGB565_TO_RED(_val)   ((((_val) & 0xf800) >> 11) << 3)
#define RGB565_TO_GREEN(_val) ((((_val) & 0x07e0) >>  5) << 2)
#define RGB565_TO_BLUE(_val)  ((((_val) & 0x001f) >>  0) << 3)
#define RGB_TO_RGB565(_r, _g, _b)  (((((_r) & 0xf8) >> 3) << 11) | ((((_g) & 0xfc) >> 2) << 5) | ((((_b) & 0xf8) >> 3) << 0))

/*
 * "Scale" the color component according with a factor. Example:
 *
 *    IN   FACTOR   OUT
 *    ---  ------   ---
 *    255   128     255     White in, 50% transparency, white out
 *      0   128     128     Black in, 50% transparency, 50% out
 *     75   128     165
 *    200   128     228
 *    200     8     253
 *    200   248     202
 */
#define PIX_SCALE(_pix, _factor)  (255 - (((255 - (_pix)) * (_factor)) >> 8))


#define ARTIST_WIDTH    270
#define ARTIST_HEIGHT    12
#define TITLE_WIDTH     270
#define TITLE_HEIGHT     12


//#define GUI_DBG(x)  x
#define GUI_DBG(x)


/******************************************************************************
 * Local Variables
 *****************************************************************************/

static displayinfo_t DisplayInfo;
static SWIM_WINDOW_T MainWindow;

static clickinfo_t ClickInfo;

static time_t oldTime = {2,2,2,142};
static image_t imgBackground = {NULL,0,0};
static image_t imgCursor = {NULL,0,0};

/*
 * Variables for drawing the digits
 */
static int duration;
static image_t imgDigits[3][10];
static point_t posDigits[3] = {{35,116}, {65,116}, {90,116}};
static image_t imgColon;
static point_t posColon = {56,116};

/*
 * Variables for drawing VU data
 */
static SWIM_WINDOW_T VuWin;
static uint16_t vuData[2 * VU_BAR_WIDTH * VU_BAR_HEIGHT]; //2 complete bars, 7 pixels wide, 32 pixels high
static image_t vuImages[VU_LEVELS];
static vu_data_t oldVuData, maxVuData;

/*
 * Variables for drawing title and artist
 */
static image_t imgTitle;
static image_t imgArtist;
static point_t posTitle = {35,78};
static point_t posArtist = {35,61};

/*
 * Areas that the user may click on. All coordinates should be relative
 * to the image - not the display. The coordinates will be updated during
 * display initialization with the display's offset.
 */
static area_t clickable_areas[] = {
  {TRUE, 30, 212,  53, 236, EVENT_PREV,  "Prev"},       
  {TRUE, 65, 212,  88, 236, EVENT_PLAY,  "Play"},       
  {TRUE,100, 212, 123, 236, EVENT_PAUSE, "Pause"},      
  {TRUE,136, 212, 159, 236, EVENT_STOP,  "Stop"},       
  {TRUE,169, 212, 192, 236, EVENT_NEXT,  "Next"},       
};

/******************************************************************************
 * Local Functions
 *****************************************************************************/

// Copies image data from pSrc at xOff, yOff with the size of 
// pDest->width * pDest->height. pDest must have allocated data of the correct size
static void copyImagePart(image_t* pSrc, image_t* pDest, int xOff, int yOff)
{
  int dstX, dstY;
  uint16_t* pDestData = (uint16_t*)pDest->pData;
  uint16_t* pSrcData;

  for (dstY = 0; dstY < pDest->height; dstY++)
  {
    pSrcData = &((uint16_t*)pSrc->pData)[(yOff + dstY)*pSrc->width + xOff];
    for (dstX = 0; dstX < pDest->width; dstX++)
    {
      *pDestData = *pSrcData;
      pDestData++;
      pSrcData++;
    }
  }
}

static event_t GUI_TouchAt(int x, int y, int z, Bool released)
{
  //log_i("%s(%d, %d, %d) released = %d\r\n", __FUNCTION__, x, y, z, released);
  if (released)
  {
    int i;
    for (i = NUM_AREAS-1; i >= 0; i--)
    {
      if (clickable_areas[i].inUse     &&
          (x >= clickable_areas[i].x0) && 
          (x <= clickable_areas[i].x1) && 
          (y >= clickable_areas[i].y0) && 
          (y <= clickable_areas[i].y1))
      {
        log_i("Clicked %s\r\n", clickable_areas[i].descr);
        return clickable_areas[i].action;
      }
    }
  }
  return EVENT_NOTHING;
}

static void drawDigit(int arr, int digit, uint16_t x, uint16_t y)
{
  swimex_put_image_at_position(
     &MainWindow, 
     (COLOR_T*)(imgDigits[arr][digit].pData), 
     x, 
     y,
     DIGIT_WIDTH, DIGIT_HEIGHT);
}

static void drawTime(int timeInSec, int maxTime)
{
  time_t newTime;
  if (timeInSec == oldTime.decimal) 
  {
    return;
  } 
  else if (timeInSec > maxTime) 
  {
    timeInSec = maxTime;
  }
  newTime.min = (timeInSec/60)%10;
  newTime.sec_high = (timeInSec % 60) / 10;
  newTime.sec_low = timeInSec % 10;
  newTime.decimal = timeInSec;

  if (oldTime.min != newTime.min)
  {
    drawDigit(0, newTime.min, posDigits[0].x, posDigits[0].y);
  }
  if (oldTime.sec_high != newTime.sec_high)
  {
    drawDigit(1, newTime.sec_high, posDigits[1].x, posDigits[1].y);
  }
  if (oldTime.sec_low != newTime.sec_low)
  {
    drawDigit(2, newTime.sec_low, posDigits[2].x, posDigits[2].y);
  }

  /* Update Cursor */
  swimex_put_image_at_position(&MainWindow, (COLOR_T*)(imgBackground.pData + imgBackground.width*(166)*2),
      DisplayInfo.bkgOffsetX,
      DisplayInfo.bkgOffsetY+166,
      imgBackground.width, imgCursor.height);
  swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgCursor.pData,
      DisplayInfo.bkgOffsetX + CURSOR_MIN + CURSOR_POS(timeInSec, maxTime),
      DisplayInfo.bkgOffsetY+166,
      imgCursor.width, imgCursor.height);

  oldTime = newTime;
}

/**
 * Draw the MP3 player GUI
 */
static void drawMP3GUI(void)
{
  int i;

  swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgBackground.pData,
      DisplayInfo.bkgOffsetX,
      DisplayInfo.bkgOffsetY,
      DisplayInfo.bkgWidth,
      DisplayInfo.bkgHeight);

  swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgColon.pData, posColon.x, posColon.y, imgColon.width, imgColon.height);

  GUI_DBG(swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgCursor.pData, 730, 20, imgCursor.width, imgCursor.height);)

  drawTime(0, 0);

#define TEST2(_col, _area) do {\
    swim_set_pen_color(&MainWindow, (_col));\
    swim_put_line(&MainWindow, (_area).x0, (_area).y0, (_area).x1, (_area).y0); \
    swim_put_line(&MainWindow, (_area).x1, (_area).y0, (_area).x1, (_area).y1); \
    swim_put_line(&MainWindow, (_area).x1, (_area).y1, (_area).x0, (_area).y1); \
    swim_put_line(&MainWindow, (_area).x0, (_area).y1, (_area).x0, (_area).y0); \
    if ((_area).action == EVENT_NOTHING) { \
      swim_put_line(&MainWindow, (_area).x0, (_area).y0, (_area).x1, (_area).y1); \
      swim_put_line(&MainWindow, (_area).x0, (_area).y1, (_area).x1, (_area).y0); \
    } \
  } while(0)

  for (i = NUM_AREAS-1; i >= 0; i--)
  {
    GUI_DBG(TEST2(RED, clickable_areas[i]);)
  }
}

static void updatePositions(void)
{
  int i;

  DisplayInfo.bkgOffsetX = (MainWindow.xpsize - imgBackground.width)/2;
  DisplayInfo.bkgOffsetY = (MainWindow.ypsize - imgBackground.height)/2;
  DisplayInfo.bkgWidth = imgBackground.width;
  DisplayInfo.bkgHeight = imgBackground.height;

  for (i = NUM_AREAS-1; i >= 0; i--)
  {
    clickable_areas[i].x0 += DisplayInfo.bkgOffsetX;
    clickable_areas[i].x1 += DisplayInfo.bkgOffsetX;
    clickable_areas[i].y0 += DisplayInfo.bkgOffsetY;
    clickable_areas[i].y1 += DisplayInfo.bkgOffsetY;
  }

  for (i = 0; i < 3; i++)
  {
    posDigits[i].x += DisplayInfo.bkgOffsetX;
    posDigits[i].y += DisplayInfo.bkgOffsetY;
  }
  posColon.x += DisplayInfo.bkgOffsetX;
  posColon.y += DisplayInfo.bkgOffsetY;

  posTitle.x += DisplayInfo.bkgOffsetX;
  posTitle.y += DisplayInfo.bkgOffsetY;
  posArtist.x += DisplayInfo.bkgOffsetX;
  posArtist.y += DisplayInfo.bkgOffsetY;
}

static ImageError_t loadImage(char* filename, image_t* info)
{
  ImageError_t err;

  err = decodeImageFileRomFS(IMAGE_FORMAT_UNKNOWN, filename, &(info->pData), &(info->width), &(info->height));
  if (err != IMAGE_ERROR_OK) 
  {
    log_e("Failed to decode %s, error %d\r\n", filename, err);
  }
  return err;
}

static void prepareVUImages()
{
  int i;
  ImageError_t err;
  image_t imgVU;

  swim_window_open_noclear(&VuWin, 
      DisplayInfo.width, DisplayInfo.height, MainWindow.fb, 
      DisplayInfo.bkgOffsetX + VU_AREA_X0,
      DisplayInfo.bkgOffsetY + VU_AREA_Y0,
      DisplayInfo.bkgOffsetX + VU_AREA_X0 + VU_AREA_WIDTH,
      DisplayInfo.bkgOffsetY + VU_AREA_Y0 + VU_AREA_HEIGHT,
      0, RED, WHITE, BLACK);

  // Load image with all bars
  err = loadImage("/mp3/gui_vu.png", &imgVU);
  if (err != IMAGE_ERROR_OK) 
  {
    // White top of image, Black bottom of image
    memset(vuData, 0xff, VU_BAR_WIDTH * VU_BAR_HEIGHT * sizeof(uint16_t));
    memset(vuData + VU_BAR_WIDTH * VU_BAR_HEIGHT, 0x00, VU_BAR_WIDTH * VU_BAR_HEIGHT * sizeof(uint16_t));
  
    // Create image data by sliding down the white-black image
    for (i = 0; i < VU_LEVELS; i++)
    {
      vuImages[i].pData = (uint8_t*)(vuData + i * VU_BAR_WIDTH);
      vuImages[i].width = VU_BAR_WIDTH;
      vuImages[i].height = VU_BAR_HEIGHT;
    }
  }
  else
  {
    GUI_DBG(swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgVU.pData,
        0, 0, imgVU.width, imgVU.height);)

    // Extract bars from the big image
    for (i = 0; i < VU_LEVELS; i++)
    {
      vuImages[i].pData = (uint8_t*)lpc_new(2 * VU_BAR_WIDTH * VU_BAR_HEIGHT); // 16bit per pixel
      vuImages[i].width = VU_BAR_WIDTH;
      vuImages[i].height = VU_BAR_HEIGHT;

      copyImagePart(&imgVU, &vuImages[i], i*VU_BAR_WIDTH, 0);
    }
    lpc_free(imgVU.pData);
  }

  GUI_DBG
  (
    swim_set_fill_color(&MainWindow, GREEN);
    swim_put_box(&MainWindow, 
       300, 
       410, 
       300+10+320, 
       410+50);
    swim_put_box(&MainWindow, 
       270, 
       390, 
       290, 
       480);
      swimex_put_image_at_position(&MainWindow, (COLOR_T*)vuData,
          275, 400, 7, 64);

    for (i = 0; i < 32; i++)
    {
      swimex_put_image_at_position(&MainWindow, (COLOR_T*)vuImages[i].pData,
          305 + 10*i, 425, 7, 32);
    }
  )

  memset(&oldVuData, 0, sizeof(oldVuData));
  memset(&maxVuData, 0, sizeof(maxVuData));
}

/*
 * Handles 0-9 and ':'. The "alldigits" images should be black digits on white background.
 * Everything white will be transparent, black will be painted as white and any gray
 * colors will create partial transparency.
 */
static void prepareOneDigitImage(image_t* pImage, image_t* alldigits, int digit, point_t pos)
{
  int x, y;
  uint16_t* pSrcData;
  uint16_t* pDestData;

  pSrcData = (uint16_t*)(alldigits->pData + DIGIT_IMG_OFFSET(digit));
  pDestData = lpc_new(DIGIT_WIDTH * DIGIT_HEIGHT * 2); //16-bit per pixel
  pImage->pData = (uint8_t*)pDestData;
  pImage->width = DIGIT_WIDTH;
  pImage->height = DIGIT_HEIGHT;
  copyImagePart(&imgBackground, pImage, pos.x, pos.y);
  for (y = 0; y < DIGIT_HEIGHT; y++)
  {
    for (x = 0; x < DIGIT_WIDTH; x++)
    {
      // Use black pixels, ignore white pixels and blend the rest
      if (*pSrcData == BLACK)
      {
        *pDestData = WHITE;
      }
      else if (*pSrcData != WHITE)
      {
        // Use the green component of the source image as it has 6 bit resolution.
        // Note that it is not really green - just a grayscale
        uint16_t factor = RGB565_TO_GREEN(*pSrcData);
        uint16_t tmpR, tmpG, tmpB;

        tmpR = PIX_SCALE(RGB565_TO_RED(*pDestData), factor);
        tmpG = PIX_SCALE(RGB565_TO_GREEN(*pDestData), factor);
        tmpB = PIX_SCALE(RGB565_TO_BLUE(*pDestData), factor);
        *pDestData = RGB_TO_RGB565(tmpR, tmpG, tmpB);
      }
      pSrcData++;
      pDestData++;
    }
  }
}

static void prepareDigitImages(void)
{
  image_t imgAllNumbers = {NULL,0,0};
  int i, j;

  // Image with digits and a colon
  if (loadImage("/mp3/gui_num.png", &imgAllNumbers) != IMAGE_ERROR_OK) 
  {
    return;
  }
  GUI_DBG(swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgAllNumbers.pData, 700, 20, imgAllNumbers.width, imgAllNumbers.height);)

  // Add backgrounds to create transparent effect
  for (i = 0; i < 3; i++)
  {
    for (j = 0; j < 10; j++)
    {
      prepareOneDigitImage(&(imgDigits[i][j]), &imgAllNumbers, j, posDigits[i]);
      GUI_DBG(drawDigit(i, j, DisplayInfo.bkgOffsetX + (DIGIT_WIDTH+3)*j, 3 + (DIGIT_HEIGHT + 3)*i);)
    }
  }
  prepareOneDigitImage(&imgColon, &imgAllNumbers, 10, posColon);
  GUI_DBG(swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgColon.pData, 
     DisplayInfo.bkgOffsetX + (DIGIT_WIDTH+3)*10, 3 + (DIGIT_HEIGHT + 3)*2,
     DIGIT_WIDTH, DIGIT_HEIGHT);)

  lpc_free(imgAllNumbers.pData);
}

static void prepareImages(void)
{
  // Main background
  if (loadImage("/mp3/gui_bkg.png", &imgBackground) != IMAGE_ERROR_OK) 
  {
    return;
  }

  // Digits - minute, high and low seconds
  prepareDigitImages();

  // Cursor (shows position in song)
  if (loadImage("/mp3/cursor.png", &imgCursor) != IMAGE_ERROR_OK) 
  {
    return;
  }

  // Backgrounds for song artist and title
  imgArtist.pData = lpc_new(ARTIST_WIDTH * ARTIST_HEIGHT * 2); //16-bit per pixel
  imgArtist.width = ARTIST_WIDTH;
  imgArtist.height = ARTIST_HEIGHT;
  copyImagePart(&imgBackground, &imgArtist, posArtist.x, posArtist.y);
  imgTitle.pData = lpc_new(TITLE_WIDTH * TITLE_HEIGHT * 2); //16-bit per pixel
  imgTitle.width = TITLE_WIDTH;
  imgTitle.height = TITLE_HEIGHT;
  copyImagePart(&imgBackground, &imgTitle, posTitle.x, posTitle.y);

  // Now that the background and digits are loaded update all relative positions
  updatePositions();

  // VU bars
  prepareVUImages();

  // Equalizer?

}

/******************************************************************************
 * Global Functions
 *****************************************************************************/

int GUI_Init(void)
{
    /* Create a SWIM window */
    swim_window_open(&MainWindow,
        GlobalParameters.display.width,
        GlobalParameters.display.height,
        (COLOR_T*)GlobalParameters.display.framebuffer, 0, 0,
        GlobalParameters.display.width-1,
        GlobalParameters.display.height-1,
        0, WHITE, BLACK, BLACK);
    
    DisplayInfo.width = GlobalParameters.display.width;
    DisplayInfo.height = GlobalParameters.display.height;

    //drawMenu();
    prepareImages();
    
    drawMP3GUI();

  /* Stop using I2C so that the UDA1380 can initialize it again */
//   I2C_Cmd(I2C_PORT, DISABLE);
//   I2C_DeInit(I2C_PORT);

  return 0;
}

void GUI_SetVolume(uint8_t vol)
{
}

void GUI_SetCurrentSong(int id, char* title, char* artist, int lengthInSec)
{
  duration = lengthInSec;

  // clear old text by copying from background image
  swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgTitle.pData, posTitle.x, posTitle.y, TITLE_WIDTH, TITLE_HEIGHT);
  swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgArtist.pData, posArtist.x, posArtist.y, ARTIST_WIDTH, ARTIST_HEIGHT);

  swim_set_font_trasparency(&MainWindow, 0);
  swim_set_pen_color(&MainWindow, WHITE);

  swim_put_text_xy(&MainWindow, artist, posArtist.x, posArtist.y);
  swim_put_text_xy(&MainWindow, title, posTitle.x, posTitle.y);
}

void GUI_UpdateTime(uint32_t timeInSec)
{
  drawTime(timeInSec, duration);
}

event_t GUI_PollTouch(void)
{
  event_t Event = EVENT_NOTHING;
  
  LED_On(LED_GUI_TOUCH);
  
  touch_xyz(&ClickInfo.newX, &ClickInfo.newY, &ClickInfo.newZ);

  if (ClickInfo.newZ == 0 && ClickInfo.oldZ != 0) 
  {
    // ignore x and y coordinates as they are wrong when z becomes 0
    ClickInfo.oldZ = 0;
    ClickInfo.pressed = FALSE;

    Event = GUI_TouchAt(ClickInfo.oldX, ClickInfo.oldY, ClickInfo.oldZ, TRUE);
  } 
  else if ((ClickInfo.oldX != ClickInfo.newX) || 
           (ClickInfo.oldY != ClickInfo.newY) || 
           (ClickInfo.oldZ == 0 && ClickInfo.newZ != 0)) 
  {
    ClickInfo.oldX = ClickInfo.newX;
    ClickInfo.oldY = ClickInfo.newY;
    ClickInfo.oldZ = ClickInfo.newZ;
    if (!ClickInfo.pressed && ClickInfo.newZ > 0) 
    {
      ClickInfo.pressed = TRUE;
    }

    Event = GUI_TouchAt(ClickInfo.oldX, ClickInfo.oldY, ClickInfo.oldZ, FALSE);
  }
  
  LED_Off(LED_GUI_TOUCH);
  
  return Event;
}

void GUI_SetCursor(int newPos)
{
  GUI_DBG(drawDigit(0, newPos, DisplayInfo.bkgOffsetX, 440);)
}

void GUI_SetVU(vu_data_t* data)
{
  int i, pos, xoff, ch;
  uint16_t* maxChannelData;
  uint16_t* newChannelData;
  
  // first left...
  xoff = VU_BAR_OFFSET_LX;
  maxChannelData = maxVuData.left;
  newChannelData = data->left;
  for (ch = 0; ch < 2; ch++)
  {
    for (i = 0; i < GUI_NUM_SAMPLES_PER_CHANNEL; i++)
    {
      pos = MIN(newChannelData[i], 31);
      swimex_put_image_at_position(&VuWin, (COLOR_T*)vuImages[pos].pData,
          xoff + VU_BAR_SPACING*+i, VU_BAR_OFFSET_Y, VU_BAR_WIDTH, VU_BAR_HEIGHT);

      // Draw "falling max"
      if (maxChannelData[i] > pos)
      {
        pos = maxChannelData[i] - 1;

        swim_put_line(&VuWin, 
           xoff + VU_BAR_SPACING*i, 
           VU_BAR_OFFSET_Y + (31-pos), 
           xoff + VU_BAR_SPACING*i + 6, 
           VU_BAR_OFFSET_Y + (31-pos));
      }
      maxChannelData[i] = pos;
    }
    
    // ...then right
    xoff = VU_BAR_OFFSET_RX;
    maxChannelData = maxVuData.right;
    newChannelData = data->right;
  }
}

void GUI_ShowWarningMessage(void)
{
  image_t imgWarning;
  if (loadImage("/mp3/warning.png", &imgWarning) == IMAGE_ERROR_OK)
  {
    swimex_put_image_at_position(&MainWindow, (COLOR_T*)imgWarning.pData,
        DisplayInfo.bkgOffsetX,
        DisplayInfo.bkgOffsetY,
        imgWarning.width,
        imgWarning.height);
  }
  else
  {
    log_e("\r\nThe MP3 Player requires an SD card with MP3 files in the root.\r\n");
    log_e("\r\nPlease insert an SD card and then reset the board\r\n");
  }
}
