/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "id3.h"
#include "ff.h"
#include "log.h"
#include "lpc_heap.h"
#include <string.h>

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define ID3V1_PAD_CHAR(_c)  (((_c)==' ') || ((_c)=='\0'))

/******************************************************************************
 * Local Variables
 *****************************************************************************/

static MP3Info_t info;
static FIL*      pFile;

/******************************************************************************
 * Global Variables
 *****************************************************************************/


/******************************************************************************
 * Local Functions
 *****************************************************************************/

//In short, data in ID3v2 tags are stored as
//"syncsafe integers". This is so the tag info
//isn't mistaken for audio data, and attempted to
//be "played". For more info, have fun Googling it.
static int syncsafe(int i)
{
	int toReturn = 0;
	toReturn |= ((i & 0x7F000000) >> 24);
	toReturn |= ((i & 0x007F0000) >>  9);
	toReturn |= ((i & 0x00007F00) <<  6);
	toReturn |= ((i & 0x0000007F) << 21);
	return toReturn;     
}

//how much room does ID3 version 2 tag info
//take up at the beginning of this file (if any)
static int id3v2size(void)
{
  FRESULT res;
  char buffer[6] = {0};
  uint32_t nRead;
  int size = 0;

  res = f_read(pFile, buffer, 6, &nRead);
  if (res == FR_OK && nRead==6)
  {
     if( buffer[0] != 'I' || buffer[1] != 'D' || buffer[2] != '3' )
     {   
         //no tag data
         return 0;
     }
    res = f_read(pFile, (char*)(&size), 4, &nRead);
    if (res == FR_OK && nRead==4)
    {
      size = syncsafe(size);
      //"size" doesn't include the 10 byte ID3v2 header
      return size + 10;
    }
  }
  return 0;
}

/*
 * Extract title and artist information if available. If the information
 * is available then it takes up 128 bytes (returned with the size parameter).
 * If the ID3v1 tag could not be found then size 0 is returned.
 *
 * Information about ID3v1 found on http://en.wikipedia.org/wiki/ID3
 */
static void extractInfoID3v1(int* size)
{
  FRESULT res;
  char buffer[128] = {0};
  uint32_t nRead;

  *size = 0;

  res = f_lseek(pFile, pFile->fsize - 128);
  if (res == FR_OK)
  {
    res = f_read(pFile, buffer, 128, &nRead);
    if (res == FR_OK && nRead==128)
    {
			if (buffer[0] == 'T' && buffer[1] == 'A' && buffer[2] == 'G')
			{
				//found tag data
				*size = 128; 

				//title is 30 characters either space- or zero-padded following the TAG
				memcpy(info.Title, &buffer[3], 30);
				info.Title[30] = '\0';

				//artist is 30 characters either space- or zero-padded following the title
				memcpy(info.Artist, &buffer[33], 30);
				info.Artist[30] = '\0';
			}
    }
  }
}

static void analyze(char* fileName)
{
  int dataEnd = pFile->fsize;
  int dataBegin = 0;
	int v1size;

	/* Calculate size of data area. The data area lies between the optional
	 * ID3v2 header at the beginning and the optional ID3v1 header at the
	 * end of the file.
	 */
	extractInfoID3v1(&v1size);
  dataEnd -= v1size;
  dataBegin += id3v2size();
  info.DataSize = dataEnd - dataBegin;
}

/******************************************************************************
 * Global Functions
 *****************************************************************************/

void ID3_ExtractInformation(char* fileName, char* longFileName, MP3Info_t** ppInfo)
{
	FRESULT err;
	
	if (pFile == NULL)
	{
		pFile = lpc_new(sizeof(FIL));
		memset(pFile, 0, sizeof(FIL));
		info.Title[ID3_MAX_TITLE_LEN] = '\0';
		info.Artist[ID3_MAX_ARTIST_LEN] = '\0';
	}

  info.Title[0] = '\0';
  info.Artist[0] = '\0';
  info.DataSize = 0;
	*ppInfo = &info;
	
	err = f_open(pFile, fileName, FA_READ);
	if (err != FR_OK)
	{
		log_e("Failed (error %d) to open %s to extract info\r\n", err, fileName);
	}
	else
	{
		analyze(fileName);
		f_close(pFile);
	}
	
	// Default to showing the file name if the ID3 tag was missing 
	if (ID3V1_PAD_CHAR(info.Title[0]))
	{
		if (longFileName != NULL && !ID3V1_PAD_CHAR(*longFileName))
		{
			strncpy(info.Title, longFileName, ID3_MAX_TITLE_LEN);
		}
		else
		{
			memcpy(info.Title, fileName, 13);
		}
	}
	if (ID3V1_PAD_CHAR(info.Artist[0]))
	{
		if (longFileName != NULL && !ID3V1_PAD_CHAR(*longFileName))
		{
			strncpy(info.Artist, longFileName, ID3_MAX_ARTIST_LEN);
		}
		else
		{
			memcpy(info.Artist, fileName, 13);
		}
	}
}
