/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/
#include <string.h>
#include "lpc177x_8x_gpdma.h"
#include "lpc177x_8x_i2s.h"
#include "lpc177x_8x_pinsel.h"
#include "lpc177x_8x_timer.h"
#include "uda1380.h"
#include "board.h"
#include "ff.h"
#include "log.h"
#include "mp3dec.h"
#include "memreg.h"
#include "joystick.h"
#include "gui.h"
#include "cfgmem.h"
#include "lpc_heap.h"
#include "sdram.h"
#include "vu.h"
#include "id3.h"
#include "application.h"

/******************************************************************************
 * Defines
 *****************************************************************************/

/** Max buffer length */
#define DATA_SIZE			434904
#define BUFFER_SIZE			1048 * 4

/* SRAM allocation: 
 *
 * IRAM: 0x10000000 - 0x10007FFF [32 kBytes] for data, bss & stack.
 * AHB0: 0x20000000 - 0x20001FFF [ 8 kBytes] for READBUF
 * AHB0: 0x20002000 - 0x20002FFF [ 4 kBytes] for USBHost
 * AHB1: 0x20004000 - 0x20007FFF [16 kBytes] for AUDIOBUF
*/

/* USBHostLite has a low file read speed (around 9ms/8kB), it might be
 * necessary to reduce USB buffer size to 4kB to ensure smooth playback.
 * However, for the SD card the file read speed is fast enough, so the
 * buffer can be 8kB for sure.
 */
#define  READBUF_BASEADDR	0x20000000
#define  READBUF_SIZE_USB  (1024*8)
#define  READBUF_SIZE_SD  (1024*8)

#define  AUDIOBUF_BASEADDR	0x20004000
#define  AUDIOBUF_SIZE		(0x1200)	/* 4608 byte (4.5KB) */

/* Allocate 2 or 3 audio buffer to store PCM after decoding */
#define  MAX_AUDIOBUF_NUM	3
#if (MAX_AUDIOBUF_NUM < 2) || (MAX_AUDIOBUF_NUM > 3)
    #error Invalid MAX_AUDIOBUF_NUM setting.
#endif

/* Wait for DMA is completed. */
#define WAIFORDMADONE   while(I2S_DMA_Done == 0) 
/* Wait for I2S TX buffer is empty. */
#define WAIFORI2SDONE	while (((LPC_I2S->STATE>>16) & 0xF) != 0)

#define SUPPORT_SD

#define I2S_DMA_CHAN	7
#if I2S_DMA_CHAN<0 || I2S_DMA_CHAN>8
	#error Invalid DMA channel number!
#endif
#define DMA_I2S_TX_FIFO	0x400A8008

#define RENAME(x,y)  RENAME2(x,y)
#define RENAME2(x,y)	x##y 

LPC_GPDMACH_TypeDef *LPC_GPDMACH = RENAME(LPC_GPDMACH, I2S_DMA_CHAN);


/******************************************************************************
 * Typedefs
 *****************************************************************************/

enum {AUDIOBUF0=0, AUDIOBUF1=1, AUDIOBUF2};
typedef struct
{
	uint32_t BaseAddr;  /* base address for the audio buf */
	int32_t Size;		/* size of the audio buf */
	uint8_t Empty;		/* buf status, 1-> empty */
} AUDIOBUF;

typedef struct 
{
  uint8_t *readBuf;
  uint8_t *readPtr;
  int bytesLeft; 
  int outOfData; 
  int eofReached;
  uint32_t startTime;
  uint32_t endTime;
  uint32_t totalDecTime;
  uint32_t nFrames;
  uint32_t totalReadFileTime;
  uint32_t nReadFileBlkCnt;
  uint32_t pcmframe_size;
  float audioSecs;
} MP3_Playback_Info_t;


/******************************************************************************
 * Global Variables
 *****************************************************************************/

uint32_t CurrentMP3 = 0;
uint32_t PreviousMP3 = 0;
FIL CurrentFile;
MP3List_t MP3List[MP3_PLAYLIST_MAX_LEN];
uint32_t mp3_cnt = 0;

volatile uint16_t* g_LED_reg = (uint16_t*)MEMREG_BASE;
volatile uint16_t  g_LED_buf = 0xA500;

/******************************************************************************
 * Local Variables
 *****************************************************************************/

// static FATFS lFatfs[1];
static FILINFO Finfo;
// #if _USE_LFN
//   char Lfname[_MAX_LFN+1];
// #endif


/* Play from audio buf 0, then move to the next buf. */
volatile uint8_t PlayingBufIdx = AUDIOBUF0;
volatile uint8_t DecodeBufIdx = AUDIOBUF0;

HMP3Decoder hMP3Decoder;

static MP3_Playback_Info_t MP3_Playback_Info;

MP3FrameInfo mp3FrameInfo; 

AUDIOBUF audiobuf[MAX_AUDIOBUF_NUM] = {
	{AUDIOBUF_BASEADDR, -1, 1},
	{AUDIOBUF_BASEADDR+AUDIOBUF_SIZE, -1, 1},
#if MAX_AUDIOBUF_NUM==3
	{AUDIOBUF_BASEADDR+AUDIOBUF_SIZE*2, -1, 1},
#endif
};

vu_data_t vudata[MAX_AUDIOBUF_NUM];

volatile uint8_t I2S_DMA_Done = 0;
volatile uint8_t DMA_Buf_Empty = 0;
volatile uint8_t PlayBackIsPaused = 0;

static MP3Info_t* id3_info;

/******************************************************************************
 * Forward Declarations of Local Functions
 *****************************************************************************/

static playback_error_t MP3_Playback_Init(void);
static playback_error_t MP3_Playback_Play(void);
static playback_error_t MP3_Playback_Stop(void);
static playback_error_t MP3_Playback_Pause(void);
static playback_error_t MP3_Playback_Resume(void);

static void MP3_DisplayFormat(void);
static void DMA_SendData(uint32_t *buf, uint32_t size);

/******************************************************************************
 * Local Functions
 *****************************************************************************/

/* Check the file name, if MP3 file, return 0, otherwise return 1 */
static uint8_t IsMP3File (const char * fname)
{
  uint32_t len = strlen(fname);;

  if ((fname[len-1]!= '3') ||
      ((fname[len-2]!= 'P') && (fname[len-2]!= 'p')) ||
      ((fname[len-3]!= 'M') && (fname[len-3]!= 'm')) ||
      (fname[len-4]!= '.'))
  {
    return 0;
  }
  return 1;  
}

/**
  * @brief  Searchs MP3 and place in MP3List[][].
  *
  * @param  CountOnly: When set, it will not list the LFN
  *         in debug terminal when an MP3 file has been found
  * @retval Number of MP3 files found in memory.
  *
  */
static uint32_t Mem_SearchMp3 (uint8_t CountOnly)
{
	uint32_t mp3_idx = 0;
	FRESULT res;
	DIR dir;
	do
	{
		/* Open root directory */
		res = f_opendir(&dir, "/");
		if (res != FR_OK) {
			log_e("Failed to open /: %d \r\n", res);
			break;
		}

		/* Search max 255 MP3 files */
		while(mp3_idx < MP3_PLAYLIST_MAX_LEN)
		{
			res = f_readdir(&dir, &Finfo);
			if ((res != FR_OK) || !Finfo.fname[0]) break;
			if ((Finfo.fattrib & AM_DIR) == 0) /* None directory */
			{
				/* Set _USE_LFN to 1 */
				/* Check if it is MP3 files (*.mp3 or *.MP3) */
				if (IsMP3File(Finfo.fname))
				{
					if(!CountOnly)
					{
						MP3List[mp3_idx].drive = 1;//CurrentDrive;
						strncpy(MP3List[mp3_idx].fname, Finfo.fname, 13);
						MP3List[mp3_idx].longname = NULL;
#if _USE_LFN
						MP3List[mp3_idx].longname = lpc_new(strlen(Finfo.lfname));
						if (MP3List[mp3_idx].longname != NULL)
						{
							strcpy(MP3List[mp3_idx].longname, Finfo.lfname);
						}
						log_i("% 2d: %s\t\t%s\r\n", mp3_idx+1, Finfo.fname, Finfo.lfname);
#else
						log_i("% 2d: %s\r\n", mp3_idx+1, Finfo.fname);
#endif
					}
					mp3_idx++;
				}
			}
		}
	} while(0);
	return mp3_idx;    
}

// static void InitFS(void)
// {
//   FRESULT res;
// 	
// #if _USE_LFN
// 	Finfo.lfname = Lfname;
// 	Finfo.lfsize = sizeof(Lfname);
// #endif
// 	
//   res = f_mount(0, &lFatfs[0]);
//   if (res != FR_OK) {
//     log_e("Failed to mount MMC card: %d\r\n", res);
//     return;
//   }
// }

static void PlayList(void)
{
  /* Locate MP3 files */
  mp3_cnt = Mem_SearchMp3(1);
  if (mp3_cnt == 0) {
    log_e("Found no MP3 files.\r\n");
    GUI_ShowWarningMessage();
    while(1);
  }
  log_i("Found %u MP3 files to play:\r\n", mp3_cnt);
  Mem_SearchMp3(0);

  /* Init MP3 decoder */
  if ( (hMP3Decoder = MP3InitDecoder()) == NULL )
  {
    log_e("MP3 Decoder Init failed!\r\n");
    return;
  }

  statemachine_run(&MP3_Playback_Init, 
	                 &MP3_Playback_Play, 
	                 &MP3_Playback_Stop,
	                 &MP3_Playback_Pause,
	                 &MP3_Playback_Resume);
}

/*----------------------------------------------------------------------------
  MP3 functions 
 *----------------------------------------------------------------------------*/

/**
  * @brief  Add an PCM frame to audio buf after decoding.
  *
  * @param  len: Specifies the frame size in bytes.
  * @retval None.
  *
  */
void MP3_AddAudioBuf (uint32_t len)
{
  int oldDecodeIdx = DecodeBufIdx;

	/* Mark the status to not-empty which means it is available to playback. */
  audiobuf[DecodeBufIdx].Empty = 0;
  audiobuf[DecodeBufIdx].Size = len;
    
	/* Point to the next buffer */
	DecodeBufIdx ++;
	if (DecodeBufIdx == MAX_AUDIOBUF_NUM)
	{
		DecodeBufIdx = 0;
	}

  /* Process the audio data to create input for the VU graphics */
	vudata[oldDecodeIdx].channels = mp3FrameInfo.nChans;
	VU_Apply(audiobuf[oldDecodeIdx].BaseAddr, 
					 audiobuf[oldDecodeIdx].Size,
					 &(vudata[oldDecodeIdx]));
}

void MP3_GetPlayingVUData(vu_data_t* pVU)
{
  *pVU = vudata[PlayingBufIdx];
}


/* Read MP3 data from file and fill in the read buffer. */
/**
  * @brief  Read data from MP3 file and fill in the Read Buffer.
  *
  * @param  readBuf: Pointer to the start of the Read Buffer
  * @param  readPtr: Pointer to the start of the left bytes in Read Buffer
  * @param  bytesLeft: Specifies the left bytes number in Read Buffer
  * @param  bytesAlign: Specifies the pad number to make sure it is aligned to 4 bytes
  * @retval Actual number of bytes read from file.
  *
  */
int MP3_FillReadBuffer(uint8_t *readBuf, uint8_t *readPtr, uint32_t bytesLeft, uint32_t bytesAlign)
{
	uint32_t nRead;

  LED_On(LED_FS_READ);

    /* Move the left bytes from the end to the front */
	memmove(readBuf+bytesAlign, readPtr, bytesLeft);

#ifdef SUPPORT_SD
	if(MP3List[CurrentMP3].drive == 1)
	{
        f_read(&CurrentFile, (void *)(readBuf+bytesLeft+bytesAlign), READBUF_SIZE_SD-bytesLeft-bytesAlign, &nRead);
		/* zero-pad to avoid finding false sync word after last frame (from old data in readBuf) */
		if (nRead < READBUF_SIZE_SD - bytesLeft - bytesAlign)
			memset(readBuf+bytesAlign+bytesLeft+nRead, 0, READBUF_SIZE_SD-bytesAlign-bytesLeft-nRead);
	}
#endif
#ifdef SUPPORT_USB
	if(MP3List[CurrentMP3].drive == 0)
	{
		f_read(&File1, (void *)(readBuf+bytesLeft+bytesAlign), READBUF_SIZE_USB-bytesLeft-bytesAlign, &nRead);
		/* zero-pad to avoid finding false sync word after last frame (from old data in readBuf) */
		if (nRead < READBUF_SIZE_USB - bytesLeft - bytesAlign)
			memset(readBuf+bytesAlign+bytesLeft+nRead, 0, READBUF_SIZE_USB-bytesAlign-bytesLeft-nRead);
	}
#endif

  LED_Off(LED_FS_READ);

	return nRead;
}

/**
  * @brief  Decode a frame.
  *
  * @param  None.
  * @retval 0: success, 1: terminated.
  *
  */
uint8_t MP3_DecodeFrame ()
{
	uint8_t word_align, frame_decoded;    
	int nRead, offset, err;

	frame_decoded = 0;
	word_align = 0;
	nRead = 0;
	offset = 0;

	LED_On(LED_IN_DECODE);                      

	do
	{
		/* somewhat arbitrary trigger to refill buffer - should always be enough for a full frame */
		if (MP3_Playback_Info.bytesLeft < 2*MAINBUF_SIZE && !MP3_Playback_Info.eofReached) 
		{
			/* Align to 4 bytes */
			word_align = (4-(MP3_Playback_Info.bytesLeft&3)) & 3;

			Timer = 0;
			
			/* Fill read buffer */
			nRead = MP3_FillReadBuffer(MP3_Playback_Info.readBuf, MP3_Playback_Info.readPtr, MP3_Playback_Info.bytesLeft, word_align);
			MP3_Playback_Info.totalReadFileTime += Timer;
			MP3_Playback_Info.nReadFileBlkCnt++;
	
			MP3_Playback_Info.bytesLeft += nRead;
			MP3_Playback_Info.readPtr = MP3_Playback_Info.readBuf + word_align;
			if (nRead == 0)
			{
				MP3_Playback_Info.eofReached = 1;	/* end of file */
				MP3_Playback_Info.outOfData = 1;
			}
		}

		/* find start of next MP3 frame - assume EOF if no sync found */
		offset = MP3FindSyncWord(MP3_Playback_Info.readPtr, MP3_Playback_Info.bytesLeft);
		if (offset < 0) {
			MP3_Playback_Info.readPtr = MP3_Playback_Info.readBuf;
			MP3_Playback_Info.bytesLeft = 0;
			continue;
		}
		MP3_Playback_Info.readPtr += offset;
		MP3_Playback_Info.bytesLeft -= offset;

		//simple check for valid header
		if(((*(MP3_Playback_Info.readPtr+1) & 24) == 8) || 
       ((*(MP3_Playback_Info.readPtr+1) & 6) != 2) || 
       ((*(MP3_Playback_Info.readPtr+2) & 240) == 240) || 
       ((*(MP3_Playback_Info.readPtr+2) & 12) == 12) || 
       ((*(MP3_Playback_Info.readPtr+3) & 3) == 2))
		{
			MP3_Playback_Info.readPtr += 1;		//header not valid, try next one
			MP3_Playback_Info.bytesLeft -= 1;
			continue;
		}

    Timer = 0;
		err = MP3Decode(hMP3Decoder, &MP3_Playback_Info.readPtr, &MP3_Playback_Info.bytesLeft, (short *)audiobuf[DecodeBufIdx].BaseAddr, 0);
		MP3_Playback_Info.totalDecTime += Timer;
		MP3_Playback_Info.nFrames++;
		if(err == -6)
		{
			MP3_Playback_Info.readPtr += 1;
			MP3_Playback_Info.bytesLeft -= 1;
			continue;
		}

		if (err) {
			/* error occurred */
			switch (err) {
			case ERR_MP3_INDATA_UNDERFLOW:
				MP3_Playback_Info.outOfData = 1;
				break;
			case ERR_MP3_MAINDATA_UNDERFLOW:
				/* do nothing - next call to decode will provide more mainData */
				break;
			case ERR_MP3_FREE_BITRATE_SYNC:
			default:
				MP3_Playback_Info.outOfData = 1;
				break;
			}
		} else {
			/* no error */
			MP3GetLastFrameInfo(hMP3Decoder, &mp3FrameInfo);				 
			frame_decoded = 1; 
		}
	} while (!frame_decoded && !MP3_Playback_Info.outOfData);

	LED_Off(LED_IN_DECODE);                      

	if (MP3_Playback_Info.outOfData == 1) 
	{
		return 0x1;   /* Decoder terminated */
	}
	else
	{
		return 0x0;   /* Decoder success. */
	}
}

/**
  * @brief  Calculate appropriate x/y/z values in order to generate real bit rate 
  *         which is closer to the target value.
  *
  * @param  bitrate: target bit rate (in Hz)
  * @param  x, y, z: pointer to the buffer to store the calculated parameters
  * @retval None
  *
  */
void I2S_SearchPara(uint32_t bitrate, uint8_t *x, uint8_t *y, uint8_t *z)
{
	uint8_t x1, y1, z1, x_div=0, y_div=0, z_div=0;//, clksel;
	float delta, delta_tmp, delta_min;
	uint32_t pclk;
	
	/* Bit 22~23 is for I2S clock divider. */
	//clksel = (LPC_SC->PCLKSEL >> 22) & 0x03;
	pclk = PeripheralClock;//LPC17xx_GetFpclk(clksel);
	delta_min = 0xFFFF;
	delta = (float)pclk / bitrate;//(bitrate*2);

	for (z1=64;z1>0;z1--)
	{
		for (y1=254;y1>0;y1--)
		{
			x1 = ((double)(y1*z1)/delta + 0.5);
			if (2*x1 >= y1)	 /* make sure x/y < 1/2 to decrease jitter */
				continue;
			
			delta_tmp = (float)y1*z1/x1 - delta;
			if (delta_tmp < 0)
			{
				delta_tmp = -delta_tmp;
			}

      /* Fast search: find parameters which only meet given precision */
#if 0 
			if (delta_tmp < delta*0.001f)
			{
				if ((y1 % x1) == 0)
				{
					*x = x1; *y = y1; *z = z1;
					log_i("%d, %d, %d\r\n", x1, y1, z1);
					return;					
				}
			}
#else
      /* Slow search for the best x/y/z parameters which generated the most accurate bitrate */
			if (delta_tmp < delta_min)
			{
				delta_min = delta_tmp;
				x_div = x1;	y_div = y1;	z_div = z1;
			}
#endif
		}
	}

  /* Store the found parameters. */
	*x = x_div; *y = y_div; *z = z_div;

#if 0
	/* Display */
	log_i("SystemFrequency: %u, pclk: %u\r\n", SystemCoreClock, pclk);
	log_i("bitrate: %d -> %d\r\n", bitrate, ((uint32_t)pclk/(2*y_div*z_div))*x_div);
	log_i("x, y, z=:%d, %d, %d\r\n", x_div, y_div, z_div);
#endif
}

/**
  * @brief  Configure I2S bit rate with specified parameters.
  *
  * @param  channel: Specifies the channel number (1, 2)
  * @param  wordwidth: Specifies data width (8, 16, 32)
  * @param  samplerate: Specifies the sample rate (in Hz)
  * @retval 0-OK, others-failure code
  *
  */
uint8_t I2S_ConfigBitrate (uint8_t channel, uint8_t wordwidth, uint32_t samplerate)
{
    uint32_t ws_half, bitrate;
    uint8_t x, y, z;
    
//  log_i("I2S_ConfigBitrate: channel = %u, wordwidth = %u, samplerate = %u\r\n", channel, wordwidth, samplerate);
    /* check parameter */
    if ((channel !=1 && channel != 2) |
        (wordwidth != 8 && wordwidth != 16) | /* 8/16/32 */
        (samplerate > 96000)) /* supported bitrate: 16-96kHz (for lpc1700) */
		{
			return (0x1); /* unsupported format */
		}
    
    I2S_Stop(LPC_I2S, I2S_TX_MODE); 
    
    /* Configure DAO in stereo or mono mode */
    LPC_I2S->DAO &= ~(1 << 2);
    if (channel == 1)
    {
      LPC_I2S->DAO |=  (1 << 2);
    }
    
    /* bit rate config */
    if (wordwidth == 8) ws_half = 16;	/* for 8 bit, set ws_halfperiod to 16 CLK */
    else                ws_half = 17;	/* for 16 bit, set ws_halfperiod to 17 CLK	*/        

    /* Assume stereo mode because the I2S peripheral outputs zeroes for the 
     * unused channel when running mono */
    bitrate = ws_half*2*samplerate;
    I2S_SearchPara(bitrate, &x, &y, &z);
    
    LPC_I2S->TXRATE = (y & 0xFF) |     	 /* Y */
                     ((x & 0xFF) << 8);	 /* X */
    LPC_I2S->TXBITRATE = z-1;            /* Z */
    
    return (0x0);    
}


/**
  * @brief  Playback MP3 file.
  *
  * @param  None
  * @retval 0: success, none-zero: failure code
  *         1: failed to decode the first frame
  *         2: Unsupported MP3 format
  *
  */
playback_error_t MP3_Playback_Init (void)
{
  FRESULT res;
  int duration;
	
  /* Init variables */
	MP3_Playback_Info.bytesLeft = 0;
	MP3_Playback_Info.outOfData = 0;
	MP3_Playback_Info.eofReached = 0;
	MP3_Playback_Info.readPtr = MP3_Playback_Info.readBuf = (uint8_t *)(READBUF_BASEADDR);
	MP3_Playback_Info.totalDecTime = 0;
	MP3_Playback_Info.nFrames = 0;
	MP3_Playback_Info.nReadFileBlkCnt = 0;
	MP3_Playback_Info.totalReadFileTime = 0;
	MP3_Playback_Info.pcmframe_size = 0;

	ID3_ExtractInformation(MP3List[CurrentMP3].fname, MP3List[CurrentMP3].longname, &id3_info);
	log_i("\r\n%s ", id3_info->Title);

	res = f_open(&CurrentFile, MP3List[CurrentMP3].fname, FA_READ);
	if (res != FR_OK) {
		log_e("could not be opened, got error %d\r\n", res);
		return PLAYBACK_ERROR_FOPEN;
	}

	log_i("is playing...\r\n");
	actualTime = 0;


	/* Decode and playback from audio buf 0 */
	DecodeBufIdx = PlayingBufIdx = AUDIOBUF0;    
	audiobuf[AUDIOBUF0].Empty = audiobuf[AUDIOBUF1].Empty = 1;                 
	if (MAX_AUDIOBUF_NUM == 3)
	{
		audiobuf[AUDIOBUF2].Empty = 1;
	}

	/* Decode the first frame to get the frame format */
	if (MP3_DecodeFrame() == 0)
	{
		MP3_Playback_Info.pcmframe_size = (mp3FrameInfo.bitsPerSample / 8) * mp3FrameInfo.outputSamps;            
		MP3_AddAudioBuf (MP3_Playback_Info.pcmframe_size);
	}
	else
	{
		log_e("\n\rFailed to decode the 1st frame, try to increase READBUF_SIZE.\n\r");
		return PLAYBACK_ERROR_DECODER_FAILED;//0x1;  /* fail to decode the first frame */
	}

	/* Second part of duration calculation - taking bitrate into account */
	duration = (id3_info->DataSize * 8) / mp3FrameInfo.bitrate;
	GUI_SetCurrentSong(CurrentMP3, id3_info->Title, id3_info->Artist, duration);
	VU_SetSampleRate(mp3FrameInfo.samprate);

	/* Display the frame format */
	MP3_DisplayFormat();

	/* Configure appropriate bit rate for each mp3, otherwise, 
	 * the mp3 will be too fast or slow with default setting 
 	 */    
	if (I2S_ConfigBitrate(mp3FrameInfo.nChans, mp3FrameInfo.bitsPerSample, mp3FrameInfo.samprate))
	{
		log_i("Unsupported MP3 format!\n\r");
    return PLAYBACK_ERROR_UNSUPPORTED_MP3;//0x2; /* Unsupported MP3 format */
	}

	/* Start I2S+DMA to send the first decoded frame */
	I2S_Start(LPC_I2S);
	DMA_SendData((uint32_t *)audiobuf[AUDIOBUF0].BaseAddr, audiobuf[AUDIOBUF0].Size);
	I2S_DMA_Done = 0;
	DMA_Buf_Empty = 0;

  return PLAYBACK_ERROR_OK;//0;
}

playback_error_t MP3_Playback_Play (void)
{
  /* Check if any free buffer space */
  if (audiobuf[DecodeBufIdx].Empty == 1)
  {
    /* Decode one frame */
    if (MP3_DecodeFrame() == 0)    
    {
			MP3_Playback_Info.pcmframe_size = (mp3FrameInfo.bitsPerSample / 8) * mp3FrameInfo.outputSamps;
   		MP3_AddAudioBuf (MP3_Playback_Info.pcmframe_size);   
    }
  }
  else
  {
		/* Decoder is faster than play, all buf is full */
		LED_Toggle(LED_DEC_FASTER);
  }

  if (MP3_Playback_Info.outOfData == 1)
  {
    return PLAYBACK_ERROR_OUT_OF_DATA;
  } 
  else if (DMA_Buf_Empty == 1)
  {
    return PLAYBACK_ERROR_DMA_EMPTY;
  }
  return PLAYBACK_ERROR_OK;
}

playback_error_t MP3_Playback_Stop (void)
{
	/* Playback complete */
  WAIFORDMADONE;
	WAIFORI2SDONE;
	I2S_Stop(LPC_I2S, I2S_TX_MODE);
       
	/* Close file */
	f_close(&CurrentFile);

	/* Don't care about return value from f_close() as the file could
		 already be closed */
	
	if (MP3_Playback_Info.outOfData == 1 || (DMA_Buf_Empty==1))
	{   
#if 1
		MP3GetLastFrameInfo(hMP3Decoder, &mp3FrameInfo);
		MP3_Playback_Info.audioSecs = ((float)MP3_Playback_Info.nFrames * mp3FrameInfo.outputSamps) / ( (float)mp3FrameInfo.samprate * mp3FrameInfo.nChans);
		log_i("Statistics: \n\r");
		log_i("    Decode: %d frames, %d ms (%d ms/frame)\n\r", MP3_Playback_Info.nFrames, MP3_Playback_Info.totalDecTime, MP3_Playback_Info.totalDecTime/MP3_Playback_Info.nFrames);
		log_i("    Audio duration(min:sec): %02d:%02d\n\r", (uint32_t)MP3_Playback_Info.audioSecs/60, (uint32_t)MP3_Playback_Info.audioSecs%60);
#ifdef SUPPORT_USB
    if(MP3List[CurrentFile.fs? CurrentMP3 : PreviousMP3].drive == 0)
		{
    	log_i("    File read speed: %d ms/%dkB\n\r",  totalReadFileTime/nReadFileBlkCnt, READBUF_SIZE_USB>>10);
		}
#endif
#ifdef SUPPORT_SD
   	if(MP3List[CurrentFile.fs? CurrentMP3 : PreviousMP3].drive == 1)
		{
   	  log_i("    File read speed: %d ms/%dkB\n\r",  MP3_Playback_Info.totalReadFileTime/MP3_Playback_Info.nReadFileBlkCnt, READBUF_SIZE_SD>>10);
		}
#endif
    log_i("    Actual time: %d ms\r\n", actualTime);
#endif 
  }

  return PLAYBACK_ERROR_OK;
}

playback_error_t MP3_Playback_Pause(void)
{
	playback_error_t err;
	int i;

	// Make sure the DMA_IRQ does not switch song
	PlayBackIsPaused = 1;

  WAIFORDMADONE;
	WAIFORI2SDONE;
	I2S_Stop(LPC_I2S, I2S_TX_MODE);
	PlayBackIsPaused = 0;
	
	// Save actual time
	// keep filling the play buffer
	for (i = 0; i < MAX_AUDIOBUF_NUM; i++)
	{
	  err = MP3_Playback_Play();
		if (err != PLAYBACK_ERROR_OK)
		{
			return err;
		}
	}
	return PLAYBACK_ERROR_OK;
}

playback_error_t MP3_Playback_Resume(void)
{	
  I2S_Start(LPC_I2S);
	
	// Fill DMA again (PlayingBufIdx was increased when pausing)
	DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
	I2S_DMA_Done = 0;	
	DMA_Buf_Empty = 0;
	
	return PLAYBACK_ERROR_OK;
}

/**
  * @brief  Displays the format information in structure mp3FrameInfo.
  *
  * @param  None
  * @retval None 
  *
  */
void MP3_DisplayFormat(void)
{
	log_i("\n\rmp3FrameInfo:\n\r");
	log_i("    bitrate: %d bps\r\n", mp3FrameInfo.bitrate);
	log_i("    nChans: %d \r\n", mp3FrameInfo.nChans);
	log_i("    samprate: %d \r\n", mp3FrameInfo.samprate);
	log_i("    bitsPerSample: %d \r\n", mp3FrameInfo.bitsPerSample);
	log_i("    outputSamples: %d \r\n", mp3FrameInfo.outputSamps);
	log_i("    layer: %d \r\n", mp3FrameInfo.layer);
	log_i("    version: %d \r\n", mp3FrameInfo.version);    

	log_i("    Audiobuff size: %d\r\n", audiobuf[AUDIOBUF0].Size);
}

/**
  * @brief  Initializes GPDMA.
  *
  * @param  None
  * @retval None
  *
  */
void DMA_Init(void)
{
  GPDMA_Channel_CFG_Type cfg;
  I2S_DMAConf_Type i2s_dma;

  GPDMA_Init();

  cfg.ChannelNum    = I2S_DMA_CHAN;
  cfg.TransferSize  = 0; //set when data is ready
  cfg.TransferWidth = 0; //set when data is ready
  cfg.SrcMemAddr    = 0; //set when data is ready
  cfg.DstMemAddr    = DMA_I2S_TX_FIFO;
  cfg.TransferType  = GPDMA_TRANSFERTYPE_M2P;
  cfg.SrcConn       = 0; //not used for M2P
  cfg.DstConn       = GPDMA_CONN_I2S_Channel_0;
  cfg.DMALLI        = 0; //not used for M2P

  GPDMA_Setup(&cfg);

  // GPDMA_Setup unconditionally set SBSize and DBSize to 32bit, which must be changed to 16bit
  LPC_GPDMACH->CControl &= ~(GPDMA_DMACCxControl_SBSize(7) | GPDMA_DMACCxControl_DBSize(7));
  LPC_GPDMACH->CControl |= GPDMA_DMACCxControl_SBSize(GPDMA_BSIZE_16) | GPDMA_DMACCxControl_DBSize(GPDMA_BSIZE_16); 

	/* Enable I2S DMA TX */
  i2s_dma.DMAIndex = I2S_DMA_1;
  i2s_dma.depth    = 4;
  I2S_DMAConfig(LPC_I2S, &i2s_dma, I2S_TX_MODE);
  I2S_DMACmd(LPC_I2S, I2S_DMA_1, I2S_TX_MODE, ENABLE);

	NVIC_EnableIRQ(DMA_IRQn);
	NVIC_SetPriority (DMA_IRQn, 1);
}

/**
  * @brief  Start GPDMA to transer data from memory to peripheral.
  *
  * @param  buf: Pointer to the buffer
  * @param  size: data length in byte
  * @retval None
  *
  */
void DMA_SendData(uint32_t *buf, uint32_t size)
{
  LED_Toggle(LED_SEND);
	LPC_GPDMACH->CSrcAddr = (uint32_t)buf;
	LPC_GPDMACH->CControl |= (size/4) & 0x0FFF; /* burst size=4 bytes */
	LPC_GPDMACH->CConfig |= 0x01;               /* Start DMA to send data */	
}

/**
  * @brief  DMA ISR.
  *
  * @param  None
  * @retval None
  *
  */
void DMA_IRQHandler(void) 
{
  uint32_t regVal;

	regVal = LPC_GPDMA->IntErrStat;
	if (regVal)
	{
		LPC_GPDMA->IntErrClr |= regVal;
	}

  regVal = LPC_GPDMA->IntTCStat;
  if (regVal)
  {
    LPC_GPDMA->IntTCClear |= regVal;

	  if (regVal & (1<<I2S_DMA_CHAN))
	  {
		  I2S_DMA_Done = 1;
			LED_On(LED_FRAME_OUT);

			/* Data in audiobuf[PlayingBufIdx] has been sent out */
			audiobuf[PlayingBufIdx].Empty = 1;
			audiobuf[PlayingBufIdx].Size = -1;
			
			/* Send the data in next audio buffer */
			PlayingBufIdx++;
			if (PlayingBufIdx == MAX_AUDIOBUF_NUM)
			{
				PlayingBufIdx = 0;
			}
				
#if 1
			if (audiobuf[PlayingBufIdx].Empty == 1)
			{
				/* If empty==1, it means read file+decoder is slower than playback
				 * (it will cause noise) or playback is over (it is ok). 
				 */
				log_i("\n\r**Play buf(%d) is empty!\n\r", PlayingBufIdx);

				DMA_Buf_Empty = 1;
				LED_Off(LED_FRAME_OUT);
				return;
			}
#endif

			/* When paused the buffer will be sent upon resume instead */
		  if (!PlayBackIsPaused)
			{
				DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
				I2S_DMA_Done = 0;	
			}
			LED_Off(LED_FRAME_OUT);
	  }
  }
}

/******************************************************************************
 * Main Functions
 *****************************************************************************/

void mp3app(void) 
{

//   TIM_TIMERCFG_Type timerCfg;
  /* 		This code will set I2S in MASTER mode, 44100Hz, Mono,16bit
   * 		This example is used to test I2S transmit mode. There is an audio
   * 		data in audiodata.c in mono 44100Hz 16 bit, it will be send out to
   * 		I2S port. User must plug Ext I2S DAC IC to hear sound.
   */
	I2S_MODEConf_Type I2S_ClkConfig;
	I2S_CFG_Type I2S_ConfigStruct;
  uint32_t i;

  GUI_Init();
  VU_Init();
  
	/* Enable I2S in the PCONP register. I2S is disabled on reset. */
	LPC_SC->PCONP |= (1 << 27);

  /* Initialize I2S peripheral ------------------------------------*/
	/* Pin configuration:
	 * Assign:	- P0.7 as I2STX_CLK
	 *    			- P0.8 as I2STX_WS
	 * 		    	- P0.9 as I2STX_SDA
	 */
	PINSEL_ConfigPin(0,7,1);
	PINSEL_ConfigPin(0,8,1);
	PINSEL_ConfigPin(0,9,1);

	I2S_Init(LPC_I2S);

	/* setup:
	 * 		- wordwidth: 16 bits
	 * 		- stereo mode
	 * 		- master mode for I2S_TX
	 * 		- Frequency = 44.1 kHz
	 */

	/* Audio Config*/
	I2S_ConfigStruct.wordwidth = I2S_WORDWIDTH_16;
	I2S_ConfigStruct.mono = I2S_STEREO;;
	I2S_ConfigStruct.stop = I2S_STOP_ENABLE;
	I2S_ConfigStruct.reset = I2S_RESET_ENABLE;
	I2S_ConfigStruct.ws_sel = I2S_MASTER_MODE;
	I2S_ConfigStruct.mute = I2S_MUTE_DISABLE;
	I2S_Config(LPC_I2S,I2S_TX_MODE,&I2S_ConfigStruct);

  LPC_I2S->DAO &= ~(0x1ff << 6);
  LPC_I2S->DAO |=  (0x010 << 6);

	/* Clock Mode Config*/
	I2S_ClkConfig.clksel = I2S_CLKSEL_FRDCLK;
	I2S_ClkConfig.fpin = I2S_4PIN_DISABLE;
	I2S_ClkConfig.mcena = I2S_MCLK_ENABLE;
	I2S_ModeConfig(LPC_I2S,&I2S_ClkConfig,I2S_TX_MODE);

	I2S_FreqConfig(LPC_I2S, 44100, I2S_TX_MODE);

	I2S_Stop(LPC_I2S, I2S_TX_MODE);

	for(i = 0; i <0x1000000; i++);

  // Important that I2S is started before UDA1380 is initialized as the
  // I2S clock must be running.
	I2S_Start(LPC_I2S);

	for(i = 0; i <0x1000000; i++);
  
	Uda1380_Init(200000, 44100);
  Uda1380_MasterVolCtrl(0, 0);

  DMA_Init();

#if _USE_LFN
	Finfo.lfname = lpc_new((_MAX_LFN+1) * sizeof(XCHAR));
	Finfo.lfsize = _MAX_LFN;
  Finfo.lfname[0] = '\0';
#endif

  PlayList();

	while(1);
}

static void cleanup(void)
{
	// cannot quit the MP3 player application - just press reset instead
}


/******************************************************************************
 * Public Functions
 *****************************************************************************/

static AppCallback_t mp3AppCallback = {
    &mp3app,
    &cleanup,
    &statemachine_systick,
    1, // 1 ms timer
    FALSE,
};

AppCallback_t* mp3AppInit(void)
{
  return &mp3AppCallback;
}
