/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc177x_8x_adc.h"
#include "uda1380.h"
#include "board.h"
#include "ff.h"
#include "diskio.h"
#include "log.h"
#include "joystick.h"
#include "gui.h"
#include "statemachine.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

void MP3_GetPlayingVUData(vu_data_t* pVU);

/******************************************************************************
 * Local Variables
 *****************************************************************************/

static          uint32_t CurrentVolume = 0x00; // Max
static volatile uint32_t NewVolume     = 0x00; // Max
static volatile uint32_t ADC_Value     = 0x00; // Max

static int Cursor = -1;

static volatile int flag_UpdateTime = 0;
static volatile int flag_CheckADC = 0;
static volatile int flag_CheckJoy = 0;
static volatile int flag_CheckTouch = 0;
static volatile int flag_UpdateVU = 0;

static playbackFunc mp3InitFunc;
static playbackFunc mp3PlayFunc;
static playbackFunc mp3StopFunc;
static playbackFunc mp3PauseFunc;
static playbackFunc mp3ResumeFunc;

static vu_data_t vu_test_data = {
  {0x0000, 0x2000, 0x4000, 0x6000, 0x8000, 0xA000, 0xC000, 0xE000},
  {0xD000, 0xB000, 0x9000, 0x7000, 0x5000, 0x3000, 0x1000, 0xF000},
  2
};

/******************************************************************************
 * Global Variables
 *****************************************************************************/

state_t CurrentState = STATE_IDLE;
volatile uint32_t Timer;  /* Performance timer (1kHz increment) */
volatile int actualTime = 0;


/******************************************************************************
 * Local Functions
 *****************************************************************************/

static void handleEvent(event_t event, uint32_t data)
{
  int NewCursor = Cursor;

  /* Update cursor */
  switch (event)
  {
  case EVENT_PREV:
  case EVENT_CURSOR_UP:
    NewCursor = (Cursor + mp3_cnt - 1) % mp3_cnt;
    break;

  case EVENT_NEXT:
  case EVENT_CURSOR_DOWN:
    NewCursor = (Cursor + 1) % mp3_cnt;
    break;

  case EVENT_CURSOR_SET:
    NewCursor = data % mp3_cnt;
    break;

  default:
    break;
  }  
  if (NewCursor != Cursor)
  {
    Cursor = NewCursor;
    GUI_SetCursor(Cursor);
  }


  /* Handle remaining events now that the cursor is correct */
  switch (event)
  {
  case EVENT_NEXT:
  case EVENT_PREV:
  case EVENT_CURSOR_CLICK:
    f_close(&CurrentFile);
    PreviousMP3 = CurrentMP3;
    CurrentMP3 = Cursor;
    break;
	
  case EVENT_PLAY:
		/* Play should only be available when idle */
		if (CurrentState == STATE_IDLE)
		{
			PreviousMP3 = CurrentMP3;
			CurrentMP3 = Cursor;
		}
    break;
	
	case EVENT_PAUSE:
	case EVENT_STOP:
  case EVENT_CURSOR_UP:
  case EVENT_CURSOR_DOWN:
  case EVENT_CURSOR_SET:
    // already handled
    break;
  
  default:
    log_e("Unsupported event %d\r\n", event);
    break;
  }  

  //log_i("Cursor: %d, Current: %d, Previous: %d\r\n", Cursor, CurrentMP3, PreviousMP3);
}

static event_t checkInputs(void)
{
  static uint8_t  last_joy = 0;
  event_t  event = EVENT_NOTHING;
  
  if (flag_UpdateTime && CurrentState == STATE_PLAYING)
  {
    GUI_UpdateTime(actualTime/1000);
    flag_UpdateTime = 0;
  }
  
  if (flag_CheckADC)
  {
    // Start conversion
    ADC_StartCmd(LPC_ADC, ADC_START_NOW);

    // Wait for conversion to complete
    while (!ADC_ChannelGetStatus(LPC_ADC, ADC_CH_TRIMPOT, ADC_DATA_DONE));

    // Read the 12 bit value, but volume is only 8 bits.
    ADC_Value = ADC_ChannelGetData(LPC_ADC, ADC_CH_TRIMPOT);
    NewVolume = (ADC_Value >> 4) & 0xfff;

    // Filter out the lowest 3 bits to avoid unnecessary volume changes
    NewVolume = NewVolume & 0xf8;
    if (NewVolume != CurrentVolume)
    {
      //log_i("Curr = %d (0x%x), New = %d (0x%x), actual = %d (0x%x)\r\n", CurrentVolume, CurrentVolume, NewVolume, NewVolume, ADC_Value, ADC_Value);
      CurrentVolume = NewVolume;
      Uda1380_MasterVolCtrl(CurrentVolume, CurrentVolume);
    }

    // Reset to receive a new value
    flag_CheckADC = 0;
  }
  if (flag_CheckJoy)
  {
    uint8_t state = joystick_read();

    if ((last_joy & JOYSTICK_DOWN) && !(state & JOYSTICK_DOWN))
    {
      event = EVENT_CURSOR_DOWN;
    }
    else if ((last_joy & JOYSTICK_UP) && !(state & JOYSTICK_UP))
    {
      event = EVENT_CURSOR_UP;
    }
    else if ((last_joy & JOYSTICK_CENTER) && !(state & JOYSTICK_CENTER))
    {
      event = EVENT_CURSOR_CLICK;
    }
    else if ((last_joy & JOYSTICK_LEFT) && !(state & JOYSTICK_LEFT))
    {
      event = EVENT_PREV;
    }
    else if ((last_joy & JOYSTICK_RIGHT) && !(state & JOYSTICK_RIGHT))
    {
      event = EVENT_NEXT;
    }

    last_joy = state;
    flag_CheckJoy = 0;
  }
  if (flag_CheckTouch && (event == EVENT_NOTHING))
  {
    event = GUI_PollTouch();
    flag_CheckTouch = 0;
  }
  if (flag_UpdateVU && CurrentState == STATE_PLAYING) 
  {
//    int i;
//    for (i = 0; i < GUI_NUM_SAMPLES_PER_CHANNEL; i++)
//    {
//      vu_test_data.left[i] = (vu_test_data.left[i] + 0x1000);
//      vu_test_data.right[i] = (vu_test_data.right[i] - (1+i)*0x800);
//    }
    MP3_GetPlayingVUData(&vu_test_data);
//    log_i("\r\nLeft:  %d, %d, %d, %d, %d, %d, %d, %d\r\n", 
//        vu_test_data.left[0], vu_test_data.left[1], vu_test_data.left[2], vu_test_data.left[3], 
//        vu_test_data.left[4], vu_test_data.left[5], vu_test_data.left[6], vu_test_data.left[7]); 
//    log_i("Right: %d, %d, %d, %d, %d, %d, %d, %d\r\n", 
//        vu_test_data.right[0], vu_test_data.right[1], vu_test_data.right[2], vu_test_data.right[3], 
//        vu_test_data.right[4], vu_test_data.right[5], vu_test_data.right[6], vu_test_data.right[7]); 
    LED_On(LED_VU_DRAW);
    GUI_SetVU(&vu_test_data);
    LED_Off(LED_VU_DRAW);
    flag_UpdateVU = 0;
  }
  return event;
}

static state_t stateIdle(void)
{
  event_t Event;
  while (TRUE)
  {
    Event = checkInputs();
    if (Event != EVENT_NOTHING)
    {
      handleEvent(Event, 0);
      switch (Event)
      {
        case EVENT_CURSOR_CLICK:
        case EVENT_PLAY:
        case EVENT_NEXT:
        case EVENT_PREV:
          return STATE_PREPARING;
      
				case EVENT_STOP:					
				case EVENT_PAUSE:					
					break;
        
        // The rest of the events are ignored
        default:
          break;
      }
    }
  }
}

static state_t statePreparing(void)
{
  playback_error_t err;
  //This is a quick one-time-per-song state and it will not handle any events
  
  //decode first frame and add first buffer to "audiobuf"
  err = mp3InitFunc();
  if (err != PLAYBACK_ERROR_OK)
  {
    log_e("Got error %d in %s\r\n", err, __FUNCTION__);
    return STATE_IDLE;
  }
  
  return STATE_PLAYING;
}

static state_t statePlaying(void)
{
  event_t Event;
  playback_error_t err;
  while (TRUE)
  {
    // check if buffer is full, otherwise add a frame and continue
    err = mp3PlayFunc();
    if ((err == PLAYBACK_ERROR_OUT_OF_DATA) || (err == PLAYBACK_ERROR_DMA_EMPTY))
    {
      Event = EVENT_NEXT;
    }
    else
    {
      Event = checkInputs();
    }
    if (Event != EVENT_NOTHING)
    {
      handleEvent(Event, 0);
      switch (Event)
      {
        case EVENT_CURSOR_CLICK:
        case EVENT_NEXT:
        case EVENT_PREV:
          return STATE_STOPPING_FOR_NEXT;
          
        case EVENT_STOP:
          return STATE_STOPPING;
				
        case EVENT_PAUSE:
          return STATE_PAUSED;
				
        // The rest of the events are ignored
        default:
          break;
      }
    }
  }
}

static state_t statePaused(void)
{
  event_t Event;
  playback_error_t err;
	
	err = mp3PauseFunc();
	if (err != PLAYBACK_ERROR_OK)
	{
		log_i("Failed (error %d) to pause, stopping\r\n", err);
		return STATE_STOPPING;
	}
	
  while (TRUE)
  {
    Event = checkInputs();
    if (Event != EVENT_NOTHING)
    {
      handleEvent(Event, 0);
      switch (Event)
      {
        case EVENT_CURSOR_CLICK:
        case EVENT_NEXT:
        case EVENT_PREV:
          return STATE_STOPPING_FOR_NEXT;
				
        case EVENT_PLAY:
        case EVENT_PAUSE:
					err = mp3ResumeFunc();
					if (err != PLAYBACK_ERROR_OK)
					{
						log_i("Failed (error %d) to resume, stopping\r\n", err);
						return STATE_STOPPING;
					}
          return STATE_PLAYING;
      
        case EVENT_STOP:
          return STATE_STOPPING;
				
        // The rest of the events are ignored
        default:
          break;
      }
    }
  }
}

static state_t stateStopping(void)
{
  //This is a quick one-time-per-song state and it will not handle any events

  // wait for DMA and I2S t complete, close file and go to STATE_IDLE
  mp3StopFunc();
  
  // Reset song position
  GUI_UpdateTime(0);
  
  return STATE_IDLE;
}

static state_t stateStoppingForNext(void)
{
  //This is a quick one-time-per-song state and it will not handle any events
  
  // wait for DMA and I2S t complete, close file and go to STATE_PREPARING
  mp3StopFunc();

  // Reset song position
  GUI_UpdateTime(0);
  
  return STATE_PREPARING;
}

/******************************************************************************
 * Global Functions
 *****************************************************************************/

void statemachine_systick(void)
{
//   static uint32_t div_ms = 0;
//   static uint32_t div_led = 0;
  static uint32_t div_adc = 0;
  static uint32_t div_joy = 0;
  static uint32_t div_gui = 0;
  static uint32_t div_touch = 0;

  Timer++;
	
	if (CurrentState != STATE_PAUSED)
	{
		actualTime++;
	}

//   if (++div_ms >= 10) 
//   {
//     div_ms = 0;
//     disk_timerproc();
//   }
  if (++div_gui >= 100) 
  {
    div_gui = 0;
    flag_UpdateTime = 1;
    flag_UpdateVU = 1;
  }
//   if (++div_led >= 200) 
//   {
//     div_led = 0;
//   } 
  if (++div_adc >= 500) 
  {
    if (flag_CheckADC == 0) 
    {
      flag_CheckADC = 1;
      div_adc = 0;
    }
  } 
  if (++div_joy >= 10) 
  {
    flag_CheckJoy = 1;
    div_joy = 0;
  } 
  if (++div_touch >= 100) 
  {
    flag_CheckTouch = 1;
    div_touch = 0;
  }

  return;
}

void statemachine_run(playbackFunc initFunc, 
                      playbackFunc playFunc, 
                      playbackFunc stopFunc,
                      playbackFunc pauseFunc, 
										  playbackFunc resumeFunc)
{
  state_t NextState = CurrentState;

  mp3InitFunc = initFunc;
  mp3PlayFunc = playFunc;
  mp3StopFunc = stopFunc;
  mp3PauseFunc = pauseFunc;
  mp3ResumeFunc = resumeFunc;

  CurrentState = STATE_IDLE;
  handleEvent(EVENT_CURSOR_SET, 0);
  
  /*
   * Before entering the state machine, the list of MP3s must have been
   * created, ID3 tags extracted and durations calculated? GUI should 
   * be started. Stystick must be running.
   */
  while (TRUE)
  {
    switch (CurrentState)
    {
      case STATE_IDLE:
        NextState = stateIdle();
        break;
      
      case STATE_PREPARING:
        NextState = statePreparing();
        break;
      
      case STATE_PLAYING:
        NextState = statePlaying();
        break;
      
			case STATE_PAUSED:
        NextState = statePaused();
        break;
      
      case STATE_STOPPING:
        NextState = stateStopping();
        break;
            
      case STATE_STOPPING_FOR_NEXT:
        NextState = stateStoppingForNext();
        break;
      
      default:
        log_i("Invalid state %d in %s\r\n", CurrentState, __FUNCTION__);
    }
    
    if (NextState != CurrentState)
    {
      log_i("Switching from state %d to %d\r\n", CurrentState, NextState);
      CurrentState = NextState;
    }
  }
}
