/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "vu.h"
#include "gui.h"
#include "log.h"
#include "arm_math.h" 
#include "math.h" 
#include "lpc177x_8x_i2s.h"
#include "lpc_heap.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

/* Specifies length of RFFT/RIFFT Process. Supported FFT Lengths are 128, 512, 2048. */
#define FFT_LEN_REAL    512

/* flag that selects forward (ifftFlagR=0) or inverse (ifftFlagR=1) transform */
#define FFT_IFFT_FLAG_R   0

/* flag that enables (bitReverseFlag=1) or disables (bitReverseFlag=0) bit reversal of output */
#define FFT_BIT_REVERSE   1

/******************************************************************************
 * Local Variables
 *****************************************************************************/

static float32_t* real_buf;//[FFT_LEN_REAL];
static float32_t* complex_buf;//[FFT_LEN_REAL * 2];
static float32_t (*acc_buf)[FFT_LEN_REAL];

static int boundaries[GUI_NUM_SAMPLES_PER_CHANNEL] = {
                     1,// 8 * (FFT_LEN_REAL/128),
                     2,//16 * (FFT_LEN_REAL/128),
                     4,//24 * (FFT_LEN_REAL/128),
                     8,//32 * (FFT_LEN_REAL/128),
                    16,//40 * (FFT_LEN_REAL/128),
                    32,//48 * (FFT_LEN_REAL/128),
                    64,//56 * (FFT_LEN_REAL/128),
                   512,//64 * (FFT_LEN_REAL/128)
                   };
static float32_t amplitudes[32] = {
    0.126953497f,
    0.507813988f,
    1.142581472f,
    2.031255951f,
    3.173837423f,
    4.57032589f,
    6.22072135f,
    8.125023804f,
    10.28323325f,
    12.69534969f,
    15.36137313f,
    18.28130356f,
    21.45514098f,
    24.8828854f,
    28.56453681f,
    32.50009522f,
    36.68956061f,
    41.13293301f,
    45.83021239f,
    50.78139877f,
    55.98649215f,
    61.44549252f,
    67.15839988f,
    73.12521423f,
    79.34593558f,
    85.82056393f,
    92.54909926f,
    99.5315416f,
    106.7678909f,
    114.2581472f,
    122.0023106f,
    130.0003809f
};


static int shouldLog = 0;
static int callCounter = 0;

// Hanning Window   (http://www.mathworks.se/help/toolbox/signal/ref/sigwin.hannclass.html)
// must be of size FFT_LEN_REAL
static float32_t* HanningWindow;


/******************************************************************************
 * Global Variables
 *****************************************************************************/


/******************************************************************************
 * Local Functions
 *****************************************************************************/

static void log_num(char* prefix, uint16_t* data, int len)
{
  int off;
  if (shouldLog)
  {
    log_i("\r\n%s:\r\n", prefix);
    for (off = 0; off < len; off += 8)
    {
      log_i("    0x%x 0x%x 0x%x 0x%x  0x%x 0x%x 0x%x 0x%x\r\n",
            data[off+0], data[off+1], data[off+2], data[off+3],
            data[off+4], data[off+5], data[off+6], data[off+7]);
    }
  }
}

static void log_float(char* prefix, float32_t* data, int len)
{
  int off;
  if (shouldLog)
  {
    log_i("\r\n%s:\r\n", prefix);
    for (off = 0; off < len; off += 8)
    {
      log_i("    %f %f %f %f  %f %f %f %f\r\n",
            data[off+0], data[off+1], data[off+2], data[off+3],
            data[off+4], data[off+5], data[off+6], data[off+7]);
    }
  }
}

static void log_numf(char* prefix, int16_t* data, int len, int m)
{
  int off, i;
  float f[8];
  if (shouldLog)
  {
    log_i("\r\n%s:\r\n", prefix);
    for (off = 0; off < len; off += 8)
    {
      for (i = 0; i < 8; i++) 
      {
        f[i] = data[off+i];
        f[i] = f[i] / (1<<m);
      }
      log_i("    %f %f %f %f  %f %f %f %f\r\n",
            f[0], f[1], f[2], f[3], f[4], f[5], f[6], f[7]);
    }
  }
}

#define COND(_x)  do { if (off==ch) { _x; } } while(0)
#define logQ15(_msg, _data) COND(log_numf((_msg), (int16_t*)(_data), 32, 15))
#define logQ79(_msg, _data) COND(log_numf((_msg), (int16_t*)(_data), 32, 9))
#define logU16(_msg, _data) COND(log_num((_msg), (uint16_t*)(_data), 32))
#define logF32(_msg, _data) COND(log_float((_msg), (float32_t*)(_data), 32))
#define logCH(_msg, _data)  log_num((_msg), (uint16_t*)(_data), GUI_NUM_SAMPLES_PER_CHANNEL)

/******************************************************************************
 * Global Functions
 *****************************************************************************/

void VU_Init(void)
{
  int n;  
  
  // Allocate buffers for the FFT
  real_buf = lpc_new(FFT_LEN_REAL * sizeof(float32_t));
  complex_buf = lpc_new(FFT_LEN_REAL * 2* sizeof(float32_t));
  acc_buf = lpc_new(FFT_LEN_REAL * 2 * sizeof(float32_t));

  // Create table for Hanning Window   (http://www.mathworks.se/help/toolbox/signal/ref/sigwin.hannclass.html)
  HanningWindow = lpc_new(FFT_LEN_REAL * sizeof(float32_t));
  for (n = 0; n < FFT_LEN_REAL/2; n++)
  {
    HanningWindow[n] = HanningWindow[FFT_LEN_REAL - 1 - n] = 0.5 * (1 - cos((2 * PI * n) / (FFT_LEN_REAL - 1)));      
  }

  callCounter = 0;
}

void VU_SetSampleRate(int sampleRate)
{
}

void VU_Apply(uint32_t srcData, int32_t srcSize, vu_data_t* dest)
{
    arm_rfft_instance_f32        S;
    arm_cfft_radix4_instance_f32 S_CFFT; 
    int off = 0;
    int i, j;
    int ch = 0;
    int numSamplesInSrc = srcSize/2; // 16bit per sample
    int numAccumulated = 0;
    q15_t* dataIn = (q15_t*)srcData;

    //shouldLog = ((*dataIn == 0) ? 0 : 1);
    shouldLog = 0;//(++callCounter)/100;
    if (shouldLog)
    {
      callCounter = 0;
      log_i("\r\n-----\r\n");
    }

    LED_On(LED_VU_PROCESS);

    /* Reset output and temporary storage */
    for (i = 0; i < GUI_NUM_SAMPLES_PER_CHANNEL; i++)
    {
      dest->left[i] = 0;
      dest->right[i] = 0;
    }
    memset(acc_buf, 0, 2*FFT_LEN_REAL*sizeof(float32_t));
     
    /* Initialize the RFFT/RIFFT module */  
    arm_rfft_init_f32(&S, &S_CFFT, FFT_LEN_REAL, FFT_IFFT_FLAG_R, FFT_BIT_REVERSE); 

    // Mono
    if (dest->channels == 1)
    {
      /* Extract FFT_LEN_REAL channel samples from the input */
      for (off = ch; (off + FFT_LEN_REAL) < numSamplesInSrc; off+=FFT_LEN_REAL)
      {
        /* Move data a real array. real_buf is now float32 format */
        for (i = 0; i < FFT_LEN_REAL; i++)
        {
          real_buf[i] = dataIn[off + i];
          real_buf[i] = real_buf[i] / (1<<15);
        }
        if (ch == 0 && off == 0 && shouldLog)
        {
          log_float("Real 128", real_buf, FFT_LEN_REAL);
          log_i("DAO: 0x%x\r\n", LPC_I2S->DAO);
        }

        logQ15("Orig", dataIn); 
        logF32("Real", real_buf); 

        /* Apply Hanning Window */
        arm_mult_f32(real_buf, HanningWindow, real_buf, FFT_LEN_REAL);

        if (ch == 0 && off == 0 && shouldLog)
        {
          log_float("Real 128 after Hanning", real_buf, FFT_LEN_REAL); 
        }
  
        /* Process the data through the RFFT/RIFFT module. This will corrupt the
         * data in real_buf as well. The result in complex_buf is FFT_LEN_REAL
         * complex values in float32 format ( r0, q0, r1, q1, ... }
         */ 
        arm_rfft_f32(&S, real_buf, complex_buf);
        
        logF32("Complex", complex_buf);

        /* Process the data through the Complex Magnitude Module for  
         * calculating the magnitude at each bin. The input is expected to
         * be FFT_LEN_REAL complex values in float32 format and the result is 
         * FFT_LEN_REAL real values in float32 format.
         * 
         * The first values 0..(FFT_LEN_REAL/2)-1 are identical to the last
         * values (FFT_LEN_REAL/2)..FFT_LEN_REAL-1 so only half needs to be
         * calculated
         */ 
        arm_cmplx_mag_f32(complex_buf, real_buf, FFT_LEN_REAL/2);  

        logF32("Squared", real_buf);
  
        /* Add up all the real values */
        numAccumulated++;
        for (i = 0; i < FFT_LEN_REAL/2; i++)
        {
          acc_buf[ch][i] += real_buf[i];
        }

        if (ch == 0 && off == 0 && shouldLog)
        {
          log_float("Complete 128", real_buf, FFT_LEN_REAL/2);
        }
      }

      log_float("Acced 128", acc_buf[ch], FFT_LEN_REAL/2);

      /* The VU is comprised of GUI_NUM_SAMPLES_PER_CHANNEL bars per
       * channel, so the FFT_LEN_REAL samples must be grouped. This
       * is done by adding up samples. The bars are not linear so 
       * a set of precalculated boundaries are used.
       */
      for (i = 0, j = 0; i < GUI_NUM_SAMPLES_PER_CHANNEL; i++)
      {
        float32_t val = 0;
        int k;
        for (; j < boundaries[i]; j++)
        {
          // MONO so make the right channel a copy of the left channel
          val += acc_buf[ch][j];
          //dest->left[i] += acc_buf[ch][j]/numAccumulated;
          //dest->right[i] = dest->left[i];
        }
        //val = val/numAccumulated;
        for (k = 0; k < 31; k++)
        {
          if (val < amplitudes[k]) 
          {
            break;
          }
        }
        dest->left[i] = dest->right[i] = k;
      }
    }

    // Stereo
    else
    {
      /* Process one channel at a time */
      for (ch = 0; ch < dest->channels; ch++)
      {
        numAccumulated = 0;

        /* Extract FFT_LEN_REAL channel samples from the input */
        for (off = ch; (off + FFT_LEN_REAL*2) < numSamplesInSrc; off+=FFT_LEN_REAL*2)
        {
          /* Move data for one channel into a real array. real_buf is now float32 format */
          for (i = 0; i < FFT_LEN_REAL; i++)
          {
            real_buf[i] = dataIn[off + 2*i];
            real_buf[i] = real_buf[i] / (1<<15);
          }
          if (ch == 0 && off == 0 && shouldLog)
          {
            log_float("Real 128", real_buf, FFT_LEN_REAL);
            log_i("DAO: 0x%x\r\n", LPC_I2S->DAO);
          }
  
          logQ15("Orig", dataIn); 
          logF32("Real", real_buf); 
  
          /* Apply Hanning Window */
          arm_mult_f32(real_buf, HanningWindow, real_buf, FFT_LEN_REAL);
  
          if (ch == 0 && off == 0 && shouldLog)
          {
            log_float("Real 128 after Hanning", real_buf, FFT_LEN_REAL); 
          }
    
          /* Process the data through the RFFT/RIFFT module. This will corrupt the
           * data in real_buf as well. The result in complex_buf is FFT_LEN_REAL
           * complex values in float32 format ( r0, q0, r1, q1, ... }
           */ 
          arm_rfft_f32(&S, real_buf, complex_buf);
          
          logF32("Complex", complex_buf);
  
          /* Process the data through the Complex Magnitude Module for  
           * calculating the magnitude at each bin. The input is expected to
           * be FFT_LEN_REAL complex values in float32 format and the result is 
           * FFT_LEN_REAL real values in float32 format.
           * 
           * The first values 0..(FFT_LEN_REAL/2)-1 are identical to the last
           * values (FFT_LEN_REAL/2)..FFT_LEN_REAL-1 so only half needs to be
           * calculated
           */ 
          arm_cmplx_mag_f32(complex_buf, real_buf, FFT_LEN_REAL/2);  
  
          logF32("Squared", real_buf);
    
          /* Add up all the real values */
          numAccumulated++;
          for (i = 0; i < FFT_LEN_REAL/2; i++)
          {
            acc_buf[ch][i] += real_buf[i];
          }
  
          if (ch == 0 && off == 0 && shouldLog)
          {
            log_float("Complete 128", real_buf, FFT_LEN_REAL/2);
          }
        }
  
        log_float("Acced 128", acc_buf[ch], FFT_LEN_REAL/2);
  
        /* The VU is comprised of GUI_NUM_SAMPLES_PER_CHANNEL bars per
         * channel, so the FFT_LEN_REAL samples must be grouped. This
         * is done by adding up samples. The bars are not linear so 
         * a set of precalculated boundaries are used.
         */
        for (i = 0, j = 0; i < GUI_NUM_SAMPLES_PER_CHANNEL; i++)
        {
          float32_t val = 0;
          int k;
          for (; j < boundaries[i]; j++)
          {
            val += acc_buf[ch][j];
          }
          //val = val/numAccumulated;
          for (k = 0; k < 31; k++)
          {
            if (val < amplitudes[k]) 
            {
              break;
            }
          }
          if (ch == 0)
          {
            dest->left[i] = k;
          }
          else
          {
            dest->right[i] = k;
          }
        }
      }
    }
    LED_Off(LED_VU_PROCESS);

    logCH("VU left", dest->left); 
    logCH("VU right", dest->right); 
}
