package com.embeddedartists.romfs;

/*
 * $Id: Image.java,v 1.1.1.1 2005/08/19 08:57:06 rebert Exp $
 *	Build a file system image.
 */

import java.io.*;


/**
 * Image
 */
public class Image {
  public static final int LITTLE_ENDIAN = 0;
  public static final int BIG_ENDIAN = 1;

  /**
   * Construct a file system image based on the contents of
   * a file tree object.
   *
   * @param source file system
   * @param byteOrder the byte order to use
   */ 
  Image(Tree source, int byteOrder) {
    int size;
    String rootName = "/";

    targetByteOrder = byteOrder;

    offset = align32(SIZEOF_DIR_ENTRY + rootName.length() + 1);
    size = treeSize(source) + offset;
    System.out.println("size = "+size);
    bytes = new byte[size];

    for (int i=0; i < bytes.length; i++)
      bytes[i] = 0;

    //
    // Prepend root directory meta data
    //
    // __fileStart
    write32(0, offset);
    // __fileNameStart
    write32(4, SIZEOF_DIR_ENTRY);
    // __fileSize
    write32(8, dirSize(source));
    // __fileNameLength
//    bytes[10] = (byte)rootName.length();
    write16(12, rootName.length());
    // __fileFlags
//    bytes[11] = (byte)(FS_DIR_TYPE | FS_IS_LAST);
    write16(14, (FS_DIR_TYPE | FS_IS_LAST));
    // Root name
    writeString(16, rootName);

    //
    // Write the contents of the file system
    //
    processEntries(source);
  }

  /**
   * Write the image in plain binary.
   *
   * @param out the image is written to this output stream
   */
  public void writeB(OutputStream out) {
    try {
      out.write(bytes);
    }
    catch (IOException err) {
      System.err.println("Could not write to " + out);
      System.exit(1);
    }
  }

  /**
   * Write the image as a compilable C array.
   *
   * @param prefix Everything preceding the assignment operator
   * @param out Which output stream to use
   */
  public void writeC(String prefix, OutputStream out) {
    char        alphabet[] = {
      '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };
    int         first, second;

    try {
      out.write((prefix + "{\n").getBytes());

      first = (int)((bytes[0] & 0xF0) >>> 4);
      second = (int)(bytes[0] & 0x0F);
      out.write("0x".getBytes());
      out.write(alphabet[first]);
      out.write(alphabet[second]);

      for (int i=1;
           i < bytes.length;
           i++)
      {
        out.write(",".getBytes());

        if (i % 16 == 0)
          out.write("\n".getBytes());

        first = (int)((bytes[i] & 0xF0) >>> 4);
        second = (int)(bytes[i] & 0x0F);
        out.write("0x".getBytes());
        out.write(alphabet[first]);
        out.write(alphabet[second]);
      }

      out.write("};\n".getBytes());
    }
    catch (IOException err)
    {
      System.err.println("Could not write to " + out);
      System.exit(1);
    }
  }

  /**
   * Write directory entries to ROM file system image.
   * Note that meta data about the root directory should
   * be prepended to any data produced by this function.
   * 
   * @param root the root to process
   */
  private void processEntries(Tree root) {
    Tree children[] = root.getChildren();
    int myOffset;
    int nameOffset;
    byte flags;

    myOffset = offset;
    nameOffset = offset + SIZEOF_DIR_ENTRY * children.length;
    offset += align32(dirSize(root));

    for (int i=0;
         i < children.length;
         i++)
    {
      String name = children[i].getName();
      flags = 0;

      // __fileStart
      write32(myOffset, offset);
      myOffset += 4;

      // __fileNameStart
      write32(myOffset, nameOffset);
      myOffset += 4;

      // __fileSize
      if (children[i].isFile())
      {
        int size = children[i].getFileSize();
        write32(myOffset, size);
        writeBytes(offset, children[i].getFileContents());
        offset += align32(size);
        flags += FS_REG_TYPE;
      }
      else
      {
        write32(myOffset, dirSize(children[i]));
        processEntries(children[i]);
        flags += FS_DIR_TYPE;
      }
      myOffset += 4;

      // __fileNameLength
//      bytes[myOffset++] = (byte)name.length();
      write16(myOffset, name.length());
      myOffset += 2;

      // __fileFlags
      if (i + 1 == children.length)
        flags += FS_IS_LAST;

//      bytes[myOffset++] = flags;
      write16(myOffset, flags);
      myOffset += 2;

      // file name
      writeString(nameOffset, name);
      nameOffset += name.length() + 1;
    }
  }

  /**
   * Write a 32-bit integer to the specified offset.
   *
   * @param offset the integer is written to this offset
   * @param data the integer to write
   */
  private void write32(int offset, int data) {
    if (targetByteOrder == BIG_ENDIAN) {
      bytes[offset + 0] = (byte)((data & 0xFF000000) >>> 24);
      bytes[offset + 1] = (byte)((data & 0x00FF0000) >>> 16);
      bytes[offset + 2] = (byte)((data & 0x0000FF00) >>>  8);
      bytes[offset + 3] = (byte)((data & 0x000000FF));
    }
    else {
      bytes[offset + 0] = (byte)((data & 0x000000FF));
      bytes[offset + 1] = (byte)((data & 0x0000FF00) >>>  8);
      bytes[offset + 2] = (byte)((data & 0x00FF0000) >>> 16);
      bytes[offset + 3] = (byte)((data & 0xFF000000) >>> 24);
    }
  }

  /**
   * Write a 16-bit integer to the specified offset.
   *
   * @param offset the integer is written to this offset
   * @param data the integer to write
   */
  private void write16(int offset, int data) {
    if (targetByteOrder == BIG_ENDIAN) {
      bytes[offset + 0] = (byte)((data & 0xFF00) >>> 8);
      bytes[offset + 1] = (byte)(data & 0x00FF);
    }
    else {
      bytes[offset + 0] = (byte)(data & 0x00FF);
      bytes[offset + 1] = (byte)((data & 0xFF00) >>> 8);
    }
  }

  /**
   * Write a String to the specified offset.
   *
   * @param offset the string is written to this offset
   * @param data the string to write
   */
  private void writeString(int offset, String data) {
    writeBytes(offset, data.getBytes());
    bytes[offset + data.length()] = 0;
  }


  /**
   * Write a byte array to the specified offset.
   *
   * @param offset the byte array is written to this offset
   * @param data the byte array to write
   */
  private void writeBytes(int offset, byte data[]) {
    for (int i=0;
         i < data.length;
         i++)
      bytes[offset + i] = data[i];
  }


  /**
   * Calculate the image size of the directory tree
   *
   * @param root the root of the directory tree
   *
   * @return the image size
   */
  private int treeSize(Tree root) {
    Tree children[];
    int size;

    size = align32(dirSize(root));
    children = root.getChildren();

    for (int i=0;
         i < children.length;
         i++)
    {
      if (children[i].isFile())
        size += align32(children[i].getFileSize());
      else
        size += treeSize(children[i]);
    }
    return size;
  }

  /**
   *  Calculate the size of one directory.
   *
   * @param node a directory node
   *
   * @return the size of the node
   */
  private int dirSize(Tree node) {
    int size = 0;
    Tree children[] = node.getChildren();
     
    for (int i = 0;
         i < children.length;
         i++)
    {
      size += SIZEOF_DIR_ENTRY;
      size += children[i].getName().length() + 1;
    }
    return size;
  }


  /**
   * Do 32-bit alignment.
   */
  private int align32(int x) {
    int res = x + (4 - x % 4) % 4;

    return res;
  }


  //------------------------------------------------ 
  private int            targetByteOrder;
  private byte           bytes[];
  private int            offset;


  /*
        Directory entries have this structure:

        struct fsFile
        {
          unsigned __fileStart      : 32;
          unsigned __fileNameStart  : 32;
          unsigned __fileSize       : 16;
          unsigned __fileNameLength :  8;
          unsigned __fileFlags      :  8;
        };

        The ...Start values are given as offset from
        where the file system begins.
        Flags are given below.
   */
  static final int	FS_MASK_TYPE = 0x01;
  static final int	 FS_DIR_TYPE = 0x01;    // File is a directory
  static final int	 FS_REG_TYPE = 0x00;    // File is not a directory
  static final int	FS_MASK_LAST = 0x02;
  static final int	 FS_IS_LAST  = 0x02;    // Last entry in directory
  static final int	 FS_NOT_LAST = 0x00;    // Not last entry
  //--- End of flags
  static final int      SIZEOF_DIR_ENTRY = 16;
}
